__all__ = ('GuildProfile', )

from scarletio import include

from ...bases import IconSlot, IconType, Slotted
from ...color import Color
from ...core import ROLES
from ...utils import DISCORD_EPOCH_START

from .fields import (
    parse_boosts_since, parse_joined_at, parse_nick, parse_pending, parse_role_ids, parse_timed_out_until,
    put_boosts_since_into, put_joined_at_into, put_nick_into, put_pending_into, put_role_ids_into,
    put_timed_out_until_into, validate_boosts_since, validate_joined_at, validate_nick, validate_pending,
    validate_role_ids, validate_timed_out_until
)

create_partial_role_from_id = include('create_partial_role_from_id')


class GuildProfile(metaclass = Slotted):
    """
    Represents a user's profile at a guild.
    
    Attributes
    ----------
    avatar_hash : `int`
        The respective user's avatar hash at the guild in `uint128`.
    
    avatar_type : `bool`
        The respective user's avatar type at the guild.
    
    boosts_since : `None`, `datetime`
        Since when the user uses it's Nitro to boost the respective guild.
        If the user does not boost the guild then is set as `None`.
    
    joined_at : `None`, `datetime`
        The date, since the user is the member of the guild.
        If this field was not included with the initial data, then it is set to `None`.
    
    nick : `None`, `str`
        The user's nick at the guild if it has.
    
    pending : `bool`
        Whether the user has not yet passed the guild's membership screening requirements.
        Defaults to `False`.
    
    role_ids : `None`, `tuple` of ``int``
        The user's roles at the guild.
    
    timed_out_until : `None`, `datetime`
        Till when the user is timed out, and cannot interact with the guild.
    """
    __slots__ = ('boosts_since', 'joined_at', 'nick', 'pending', 'role_ids', 'timed_out_until')
    
    avatar = IconSlot('avatar', 'avatar', None, None)
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a new guild profile instance from the given parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining how should the guild profile's parameters be set.
        
        Other Parameters
        ----------------
        avatar : `None`, ``Icon``, `str`, Optional (Keyword only)
            The channel's avatar.
            
            > Mutually exclusive with `avatar_type` and `avatar_hash` parameters.
        
        avatar_type : ``IconType``, Optional (Keyword only)
            The channel's avatar's type.
            
            > Mutually exclusive with the `avatar` parameter.
        
        avatar_hash : `int`, Optional (Keyword only)
            The channel's avatar's hash.
            
            > Mutually exclusive with the `avatar` parameter.
            
        boosts_since : `None`, `datetime`, Optional (Keyword only)
            Since when the user uses it's Nitro to boost the respective guild.
        
        joined_at : `None`, `datetime`, Optional (Keyword only)
            The date, since the user is the member of the guild.
        
        nick : `None`, `str`, Optional (Keyword only)
            The user's nick at the guild if it has.
        
        pending : `bool`, Optional (Keyword only)
            Whether the user has not yet passed the guild's membership screening requirements.
        
        role_ids : `None`, `iterable` of (`int`, ``Role``), Optional (Keyword only)
            The user's roles at the guild.
        
        timed_out_until : `None`, `datetime`, Optional (Keyword only)
            Till when the user is timed out, and cannot interact with the guild.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        self = cls._create_empty()
        self._populate_from_keyword_parameters(keyword_parameters)
        return self
    
    
    def __repr__(self):
        """Returns the representation of the guild profile."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the guild profile's hash value."""
        hash_value = 0
        
        # avatar
        hash_value ^= hash(self.avatar)
        
        # boosts_since
        boosts_since = self.boosts_since
        if (boosts_since is not None):
            hash_value ^= hash(boosts_since)
        
        # joined_at
        joined_at = self.joined_at
        if (joined_at is not None):
            hash_value ^= hash(joined_at)
        
        # nick
        nick = self.nick
        if (nick is not None):
            hash_value ^= hash(nick)
        
        # pending
        hash_value ^= self.pending
        
        # role_ids
        role_ids = self.role_ids
        if (role_ids is not None):
            hash_value ^= len(role_ids) << 4
            for role_id in role_ids:
                hash_value ^= role_id
        
        # timed_out_until
        timed_out_until = self.timed_out_until
        if (timed_out_until is not None):
            hash_value ^= hash(timed_out_until)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two guild profiles are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # avatar
        if self.avatar != other.avatar:
            return False
        
        # boosts_since
        if self.boosts_since != other.boosts_since:
            return False
        
        # joined_at
        if self.joined_at != other.joined_at:
            return False
        
        # nick
        if self.nick != other.nick:
            return False
        
        # pending
        if self.pending != other.pending:
            return False
        
        # role_ids
        if self.role_ids != other.role_ids:
            return False
        
        # timed_out_until
        if self.timed_out_until != other.timed_out_until:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a guild profile from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received guild profile data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.joined_at = parse_joined_at(data)
        self._update_attributes(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the guild profile to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # avatar
        type(self).avatar.put_into(self.avatar, data, defaults, as_data = not include_internals)
        
        # boosts_since
        if include_internals:
            put_boosts_since_into(self.boosts_since, data, defaults)
        
        # joined_at
        if include_internals:
            put_joined_at_into(self.joined_at, data, defaults)
        
        # nick
        put_nick_into(self.nick, data, defaults)
        
        # pending
        if include_internals:
            put_pending_into(self.pending, data, defaults)
        
        # role_ids
        put_role_ids_into(self.role_ids, data, defaults)
        
        # timed_out_until
        put_timed_out_until_into(self.timed_out_until, data, defaults)
        
        return data
    
    
    def _set_joined(self, data):
        """
        Sets ``.joined_at`` of the guild profile if it is not set yet.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received guild profile data.
        """
        if self.joined_at is None:
            self.joined_at = parse_joined_at(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the guild profile with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received guild profile data.
        """
        self._set_avatar(data)
        self.boosts_since = parse_boosts_since(data)
        self.nick = parse_nick(data)
        self.pending = parse_pending(data)
        self.role_ids = parse_role_ids(data)
        self.timed_out_until = parse_timed_out_until(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the guild profile and returns it's changed attributes in a `dict` within `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-------------------+-------------------------------+
        | Keys              | Values                        |
        +===================+===============================+
        | avatar            | ``Icon``                      |
        +-------------------+-------------------------------+
        | boosts_since      | `None`, `datetime`            |
        +-------------------+-------------------------------+
        | nick              | `None`, `str`                 |
        +-------------------+-------------------------------+
        | pending           | `bool`                        |
        +-------------------+-------------------------------+
        | role_ids          | `None`, `tuple` of `int`      |
        +-------------------+-------------------------------+
        | timed_out_until   | `None`, `datetime`            |
        +-------------------+-------------------------------+
        """
        old_attributes = {}
        
        # avatar
        self._update_avatar(data, old_attributes)
        
        # boosts_since
        boosts_since = parse_boosts_since(data)
        if self.boosts_since != boosts_since:
            old_attributes['boosts_since'] = self.boosts_since
            self.boosts_since = boosts_since
        
        # nick
        nick = parse_nick(data)
        if self.nick != nick:
            old_attributes['nick'] = self.nick
            self.nick = nick
        
        # pending
        pending = parse_pending(data)
        if pending != self.pending:
            old_attributes['pending'] = self.pending
            self.pending = pending
        
        # role_ids
        role_ids = parse_role_ids(data)
        if role_ids != self.role_ids:
            old_attributes['role_ids'] = self.role_ids
            self.role_ids = role_ids
        
        # timed_out_until
        timed_out_until = parse_timed_out_until(data)
        if self.timed_out_until != timed_out_until:
            old_attributes['timed_out_until'] = self.timed_out_until
            self.timed_out_until = timed_out_until
        
        return old_attributes
    
    
    def copy(self):
        """
        Copies the guild profile.
        
        Returns
        -------
        new : `instance<type<self>>
        """
        new = object.__new__(type(self))
        new.avatar_hash = self.avatar_hash
        new.avatar_type = self.avatar_type
        new.boosts_since = self.boosts_since
        new.joined_at = self.joined_at
        new.nick = self.nick
        new.pending = self.pending
        role_ids = self.role_ids
        if (role_ids is not None):
            role_ids = (*role_ids,)
        new.role_ids = role_ids
        new.timed_out_until = self.timed_out_until
        return new
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the guild profile and modifies the defined the defined fields of it.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining how should the guild profile's parameters be set.
        
        Other Parameters
        ----------------
        avatar : `None`, ``Icon``, `str`, Optional (Keyword only)
            The channel's avatar.
            
            > Mutually exclusive with `avatar_type` and `avatar_hash` parameters.
        
        avatar_type : ``IconType``, Optional (Keyword only)
            The channel's avatar's type.
            
            > Mutually exclusive with the `avatar` parameter.
        
        avatar_hash : `int`, Optional (Keyword only)
            The channel's avatar's hash.
            
            > Mutually exclusive with the `avatar` parameter.
            
        boosts_since : `None`, `datetime`, Optional (Keyword only)
            Since when the user uses it's Nitro to boost the respective guild.
        
        joined_at : `None`, `datetime`, Optional (Keyword only)
            The date, since the user is the member of the guild.
        
        nick : `None`, `str`, Optional (Keyword only)
            The user's nick at the guild if it has.
        
        pending : `bool`, Optional (Keyword only)
            Whether the user has not yet passed the guild's membership screening requirements.
        
        role_ids : `None`, `iterable` of (`int`, ``Role``), Optional (Keyword only)
            The user's roles at the guild.
        
        timed_out_until : `None`, `datetime`, Optional (Keyword only)
            Till when the user is timed out, and cannot interact with the guild.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        new = self.copy()
        new._populate_from_keyword_parameters(keyword_parameters)
        return new
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates a new guild profile with it's default attributes set.
        
        Returns
        -------
        self `instance<cls>`
        """
        self = object.__new__(cls)
        self.avatar_hash = 0
        self.avatar_type = IconType.none
        self.boosts_since = None
        self.joined_at = None
        self.nick = None
        self.pending = False
        self.role_ids = None
        self.timed_out_until = None
        return self
    

    def _populate_from_keyword_parameters(self, keyword_parameters):
        """
        Sets the guild profile's attributes based on the given values.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters defining how should the guild profile's parameters be set.
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
        """
        if not keyword_parameters:
            return
        
        # avatar
        avatar = type(self).avatar.parse_from_keyword_parameters(keyword_parameters, allow_data = False)
        if (avatar is not None):
            self.avatar = avatar
        
        # boosts_since
        try:
            boosts_since = keyword_parameters.pop('boosts_since')
        except KeyError:
            pass
        else:
            self.boosts_since = validate_boosts_since(boosts_since)
        
        # joined_at
        try:
            joined_at = keyword_parameters.pop('joined_at')
        except KeyError:
            pass
        else:
            self.joined_at = validate_joined_at(joined_at)
        
        # nick
        try:
            nick = keyword_parameters.pop('nick')
        except KeyError:
            pass
        else:
            self.nick = validate_nick(nick)
        
        # pending
        try:
            pending = keyword_parameters.pop('pending')
        except KeyError:
            pass
        else:
            self.pending = validate_pending(pending)
        
        # role_ids
        try:
            role_ids = keyword_parameters.pop('role_ids')
        except KeyError:
            pass
        else:
            self.role_ids = validate_role_ids(role_ids)
        
        # timed_out_until
        try:
            timed_out_until = keyword_parameters.pop('timed_out_until')
        except KeyError:
            pass
        else:
            self.timed_out_until = validate_timed_out_until(timed_out_until)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
    
    
    def get_top_role(self, default = None):
        """
        Returns the top role of the guild profile. If the profile has no roles, then returns the `default`'s value.
        
        Parameters
        ----------
        default : `Any` = `None`, Optional
            Default value to return if the respective user has no roles at the respective guild. Defaults to `None`.
        
        Returns
        -------
        top_role : ``Role``, `default`
        """
        top_role = default
        
        role_ids = self.role_ids
        if (role_ids is not None):
            role_found = False
            
            for role_id in role_ids:
                try:
                    role = ROLES[role_id]
                except KeyError:
                    continue
                
                if role_found:
                    if role > top_role:
                        top_role = role
                else:
                    top_role = role
                    role_found = True
            
            # If non of the roles is cached, create a partial role from the highest id.
            if not role_found:
                top_role = create_partial_role_from_id(role_ids[-1])
        
        return top_role
    
    
    def iter_role_ids(self):
        """
        Iterates over the role-ids of the guild profile.
        
        This method is an iterable generator.
        
        Yields
        ------
        role_id : `int`
        """
        role_ids = self.role_ids
        if (role_ids is not None):
            yield from role_ids
    
    
    def iter_roles(self):
        """
        Iterates over the roles of the guild profile. The roles are not sorted by position like at the case of
        ``.roles``, but based on `.id`.
        
        This method is an iterable generator.
        
        Yields
        ------
        role : ``Role``
        """
        role_ids = self.role_ids
        if (role_ids is not None):
            for role_id in role_ids:
                yield create_partial_role_from_id(role_id)
    
    
    @property
    def roles(self):
        """
        Returns the roles of the guild profile in sorted form.
        
        Returns
        -------
        roles : `None`, `list` of ``Role``
        """
        role_ids = self.role_ids
        if role_ids is None:
            roles = None
        else:
            roles = sorted(create_partial_role_from_id(role_id) for role_id in self.role_ids)
        
        return roles
    
    
    @property
    def color(self):
        """
        Returns the color of the respective user at the respective guild.
        
        Returns
        -------
        color : ``Color``
        """
        role_ids = self.role_ids
        if (role_ids is not None):
            for role in sorted((create_partial_role_from_id(role_id) for role_id in self.role_ids), reverse=True):
                color = role.color
                if color:
                    return color
        
        return Color()
    
    
    @property
    def created_at(self):
        """
        Returns ``.joined_at`` if set, else the Discord epoch.
        
        Returns
        -------
        created_at : `datetime`
        """
        created_at = self.joined_at
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        
        return created_at



__all__ = ()

import reprlib, warnings
from datetime import datetime

from scarletio import Compound

from ...bases import maybe_snowflake
from ...channel import Channel, VoiceRegion
from ...exceptions import DiscordException
from ...guild import (
    AuditLog, AuditLogEvent, AuditLogIterator, COMMUNITY_FEATURES, ContentFilterLevel, Guild, GuildFeature,
    GuildPreview, GuildWidget, MessageNotificationLevel, SystemChannelFlag, VerificationLevel, VerificationScreen,
    VerificationScreenStep, WelcomeChannel, WelcomeScreen, create_partial_guild_from_data, create_partial_guild_from_id
)
from ...http import DiscordHTTPClient, VALID_ICON_MEDIA_TYPES, VALID_ICON_MEDIA_TYPES_EXTENDED
from ...localization.utils import Locale
from ...role import Role
from ...user import ClientUserBase, GuildProfile, PremiumType, User, UserBase
from ...utils import datetime_to_timestamp, get_image_media_type, image_to_base64, log_time_converter

from ..request_helpers import get_guild_and_id, get_guild_id, get_role_id, get_user_id


class ClientCompoundGuildEndpoints(Compound):
    
    guild_profiles : dict
    guilds : set
    http : DiscordHTTPClient
    id : int
    bot : bool
    premium_type : PremiumType
    
    
    async def guild_preview_get(self, guild):
        """
        Requests the preview of a public guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The id of the guild, what's preview will be requested
        
        Returns
        -------
        preview : ``GuildPreview``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = await self.http.guild_preview_get(guild_id)
        return GuildPreview(data)
    
    
    async def guild_user_delete(self, guild, user, *, reason = None):
        """
        Removes the given user from the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild from where the user will be removed.
        user : ``ClientUserBase``, `int`
            The user to delete from the guild.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `user` was not given neither as ``ClientUserBase``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        user_id = get_user_id(user)
        
        await self.http.guild_user_delete(guild_id, user_id, reason)
    
    
    async def welcome_screen_get(self, guild):
        """
        Requests the given guild's welcome screen.
        
        You need to have `manage guild` permission to request the welcome screen if the guild has it disabled.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's welcome screen will be requested.
        
        Returns
        -------
        welcome_screen : `None`, ``WelcomeScreen``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the guild has no welcome screen enabled, will not do any request.
        """
        guild_id = get_guild_id(guild)
        
        welcome_screen_data = await self.http.welcome_screen_get(guild_id)
        if welcome_screen_data is None:
            welcome_screen = None
        else:
            welcome_screen = WelcomeScreen.from_data(welcome_screen_data)
        
        return welcome_screen
    
    
    async def welcome_screen_edit(self, guild, *, enabled=..., description = ..., welcome_channels=...):
        """
        Edits the given guild's welcome screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's welcome screen will be edited.
        enabled : `bool`, Optional (Keyword only)
            Whether the guild's welcome screen should be enabled.
        description : `None`, `str`, Optional (Keyword only)
            The welcome screen's new description. It's length can be in range [0:140].
        welcome_channels : `None`, ``WelcomeChannel`` or  (`tuple`, `list`) of ``WelcomeChannel``
                , Optional (Keyword only)
            The channels mentioned on the welcome screen.
        
        Returns
        -------
        welcome_screen : `None or ``WelcomeScreen``
            The updated welcome screen. Always returns `None` if no change was propagated.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `welcome_channels` was not given neither as `None`, ``WelcomeChannel`` nor as (`tuple`, `list`) of
                ``WelcomeChannel``-s.
            - If `welcome_channels` contains a non ``WelcomeChannel``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `enabled` was not given as `bool`.
            - If `description` was not given neither as `None`, `str`.
            - If `description`'s length is out of range [0:140].
            - If `welcome_channels`'s length is out of range [0:5].
        """
        guild_id = get_guild_id(guild)
        
        data = {}
        
        if (enabled is not ...):
            if __debug__:
                if not isinstance(enabled, bool):
                    raise AssertionError(
                        f'`enabled` can be `bool`, got {enabled.__class__.__name__}; {enabled!r}.'
                    )
            
            data['enabled'] = enabled
        
        if (description is not ...):
            if __debug__:
                if (description is not None):
                    if not isinstance(description, str):
                        raise AssertionError(
                            f'`description` can be `None`, `str`, got {description.__class__.__name__}; '
                            f'{description!r}.'
                        )
                
                    description_length = len(description)
                    if description_length > 300:
                        raise AssertionError(
                            f'`description` length can be in range [0:140], got {description_length!r}; '
                            f'{description!r}.'
                        )
                
            if (description is not None) and (not description):
                description = None
            
            data['description'] = description
        
        if (welcome_channels is not ...):
            welcome_channel_datas = []
            if welcome_channel_datas is None:
                pass
            elif isinstance(welcome_channels, WelcomeChannel):
                welcome_channel_datas.append(welcome_channels.to_data())
            elif isinstance(welcome_channels, (list, tuple)):
                if __debug__:
                    welcome_channels_length = len(welcome_channels)
                    if welcome_channels > 5:
                        raise AssertionError(
                            f'`welcome_channels` length can be in range [0:5], got '
                            f'{welcome_channels_length!r}; {welcome_channels!r}.'
                        )
                
                for index, welcome_channel in enumerate(welcome_channels):
                    if not isinstance(welcome_channel, WelcomeChannel):
                        raise TypeError(
                            f'`welcome_channels[{index}]` is not `{WelcomeChannel.__name__}` , got '
                            f'{welcome_channel.__class__.__name__}; {welcome_channel!r}; '
                            f'welcome_channels={welcome_channels!r}.'
                        )
                    
                    welcome_channel_datas.append(welcome_channel.to_data())
            else:
                raise TypeError(
                    f'`welcome_channels` can be `None`, `{WelcomeChannel.__name__}`, '
                    f'(`list`, `tuple`) of `{WelcomeChannel.__name__}, got '
                    f'{welcome_channels.__class__.__name__}; {welcome_channels!r}.'
                )
            
            data['welcome_channels'] = welcome_channel_datas
        
        if data:
            data = await self.http.welcome_screen_edit(guild_id, data)
            if data:
                welcome_screen = WelcomeScreen.from_data(data)
            else:
                welcome_screen = None
        else:
            welcome_screen = None
        
        return welcome_screen
    
    
    async def verification_screen_get(self, guild):
        """
        Requests the given guild's verification screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's verification screen will be requested.

        Returns
        -------
        verification_screen : `None`, ``VerificationScreen``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the guild has no verification screen enabled, will not do any request.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        if (guild is None) or guild.has_feature(GuildFeature.verification_screen_enabled):
            verification_screen_data = await self.http.verification_screen_get(guild_id)
            if verification_screen_data is None:
                verification_screen = None
            else:
                verification_screen = VerificationScreen.from_data(verification_screen_data)
        else:
            verification_screen = None
        
        return verification_screen
    
    
    async def verification_screen_edit(self, guild, *, enabled = ..., description = ..., steps = ...):
        """
        Requests the given guild's verification screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's verification screen will be requested.
        enabled : `bool`, Optional (Keyword only)
            Whether the guild should have verification screen enabled.
        description : `None`, `str`, Optional (Keyword only)
            The guild's new description showed on the verification screen. It's length can be in range [0:300].
        steps : `None`, ``VerificationScreenStep`` or  (`tuple`, `list`) of ``VerificationScreenStep``
                , Optional (Keyword only)
            The new steps of the verification screen.
        
        Returns
        -------
        verification_screen : `None`, ``VerificationScreen``
            The updated verification screen. Always returns `None` if no change was propagated.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `steps` was not given neither as `None`, ``VerificationScreenStep`` nor as (`tuple`, `list`) of
                ``VerificationScreenStep``-s.
            - If `steps` contains a non ``VerificationScreenStep``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `enabled` was not given as `bool`.
            - If `description` was not given neither as `None`, `str`.
            - If `description`'s length is out of range [0:300].
        
        Notes
        -----
        When editing steps, `DiscordException Internal Server Error (500): 500: Internal Server Error` will be dropped.
        """
        guild_id = get_guild_id(guild)
        
        data = {}
        
        if (enabled is not ...):
            if __debug__:
                if not isinstance(enabled, bool):
                    raise AssertionError(
                        f'`enabled` can be `bool`, got {enabled.__class__.__name__}; {enabled!r}.'
                    )
            
            data['enabled'] = enabled
        
        if (description is not ...):
            if __debug__:
                if (description is not None):
                    if not isinstance(description, str):
                        raise AssertionError(
                            f'`description` can be `None`, `str`, got {description.__class__.__name__}; '
                            f'{description!r}.'
                        )
                    
                    description_length = len(description)
                    if description_length > 300:
                        raise AssertionError(
                            f'`description` length can be in range [0:300], got '
                            f'{description_length!r}; {description!r}.'
                        )
            
            if (description is not None) and (not description):
                description = None
            
            data['description'] = description
        
        if (steps is not ...):
            step_datas = []
            if steps is None:
                pass
        
            elif isinstance(steps, VerificationScreenStep):
                step_datas.append(steps.to_data())
            
            elif isinstance(steps, (list, tuple)):
                for index, step in enumerate(steps):
                    if not isinstance(step, VerificationScreenStep):
                        raise TypeError(
                            f'`steps[{index}]` is not `{VerificationScreenStep.__name__}`, got '
                            f'{step.__class__.__name__}; {step!r}; steps={steps!r}.'
                        )
                    
                    step_datas.append(step.to_data())
            else:
                raise TypeError(
                    f'`steps` can be `None`, `{VerificationScreenStep.__name__}`, '
                    f'(`list`, `tuple`) of `{VerificationScreenStep.__name__}, got '
                    f'{steps.__class__.__name__}; {steps!r}.'
                )
            
            data['form_fields'] = step_datas
        
        if data:
            data['version'] = datetime_to_timestamp(datetime.utcnow())
            data = await self.http.verification_screen_edit(guild_id, data)
            if data is None:
                verification_screen = None
            else:
                verification_screen = VerificationScreen.from_data(data)
        else:
            verification_screen = None
        
        return verification_screen
    
    
    async def guild_get(self, guild):
        """
        Gets or updates the guild.
        
        > The client must be in the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to request.
        
        Returns
        -------
        guild : ``Guild``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
        """
        guild, guild_id = get_guild_and_id(guild)
        data = await self.http.guild_get(guild_id, {'with_counts': True})
        
        if guild is None:
            channel_datas = await self.http.guild_channel_get_all(guild_id)
            data['channels'] = channel_datas
            user_data = await self.http.guild_user_get(guild_id, self.id)
            data['members'] = [user_data]
            guild = Guild.from_data(data, self)
        else:
            guild._sync(data)
        
        return guild
    
    
    async def guild_sync(self, guild):
        """
        Syncs a guild by it's id with the wrapper. Used internally if de-sync is detected when parsing dispatch events.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to sync.

        Returns
        -------
        guild : ``Guild``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # sadly guild_get does not returns channel and voice state data at least we can request the channels
        guild, guild_id = get_guild_and_id(guild)
        
        if guild is None:
            data = await self.http.guild_get(guild_id, None)
            channel_datas = await self.http.guild_channel_get_all(guild_id)
            data['channels'] = channel_datas
            user_data = await self.http.guild_user_get(guild_id, self.id)
            data['members'] = [user_data]
            guild = Guild.from_data(data, self)
        else:
            data = await self.http.guild_get(guild_id, None)
            guild._sync(data)
            channel_datas = await self.http.guild_channel_get_all(guild_id)
            guild._sync_channels(channel_datas)
            
            user_data = await self.http.guild_user_get(guild_id, self.id)
            try:
                profile = self.guild_profiles[guild.id]
            except KeyError:
                self.guild_profiles[guild.id] = GuildProfile.from_data(user_data)
                if guild not in guild.clients:
                    guild.clients.append(self)
            else:
                profile._update_attributes(user_data)
        
        return guild

##    # Disable user syncing, takes too much time
##    async def _guild_sync_post_process(self, guild):
##        for client in CLIENTS.values():
##            try:
##                user_data = await self.http.guild_user_get(guild.id, client.id)
##           except (DiscordException, ConnectionError):
##                continue
##            try:
##                profile = client.guild_profiles[guild.id]
##            except KeyError:
##                client.guild_profiles[guild.id] = GuildProfile.from_data(user_data)
##                if client not in guild.clients:
##                    guild.clients.append(client)
##            else:
##                profile._update_attributes(user_data)
##
##        if not CACHE_USER:
##            return
##
##        old_ids = set(guild.users)
##        data = {'limit': 1000, 'after': '0'}
##        while True:
##            user_datas = await self.http.guild_users(guild.id, data)
##            for user_data in user_datas:
##                user = User._create_and_update(user_data, guild)
##                try:
##                    old_ids.remove(user.id)
##                except KeyError:
##                    pass
##
##             if len(user_datas)< 1 000:
##                 break
##
##             data['after'] = user_datas[999]['user']['id']
##
##        del data
##
##        for id_ in old_ids:
##            try:
##               user = guild.users.pop(id_)
##           except KeyError:
##               continue #huh?
##           try:
##               del user.guild_profiles[guild.id]
##           except KeyError:
##               pass # huh??


    async def guild_delete(self, guild):
        """
        Deletes the given guild. The client must be the owner of the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to delete.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        await self.http.guild_delete(guild_id)
    
    
    async def guild_create(
        self, name, *, icon=None, roles=None, channels=None, afk_channel_id=None, system_channel_id=None,
        afk_timeout = None, verification_level=VerificationLevel.medium,
        message_notification=MessageNotificationLevel.only_mentions, content_filter=ContentFilterLevel.disabled,
        boost_progress_bar_enabled=None
    ):
        """
        Creates a guild with the given parameter. Bot accounts can create guilds only when they have less than 10.
        User account guild limit is 100, meanwhile staff guild limit is 200.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`
            The name of the new guild.
        icon : `None`, `bytes-like` = `None`, Optional (Keyword only)
            The icon of the new guild.
        roles : `None`, `list` of `dict` = `None`, Optional (Keyword only)
            A list of roles of the new guild. It should contain role data objects.
        channels : `None`, `list` of `dict` = `None`, Optional (Keyword only)
            A list of channels of the new guild. It should contain channel data objects.
        afk_channel_id : `None`, `int` = `None`, Optional (Keyword only)
            The id of the guild's afk channel. The id should be one of the channel's id from `channels`.
        system_channel_id: `int`, Optional (Keyword only)
            The id of the guild's system channel. The id should be one of the channel's id from `channels`.
        afk_timeout : `None`, `int` = `None`, Optional (Keyword only)
            The afk timeout for the users at the guild's afk channel.
        
        verification_level : ``VerificationLevel``, `int` = `VerificationLevel.medium`, Optional (Keyword only)
            The verification level of the new guild.
        message_notification : ``MessageNotificationLevel``, `int` = `MessageNotificationLevel.only_mentions`
                , Optional (Keyword only)
            The message notification level of the new guild.
        content_filter : ``ContentFilterLevel``, `int` = `ContentFilterLevel.disabled`, Optional (Keyword only)
            The content filter level of the guild.
        boost_progress_bar_enabled : `None`, `bool` = `None`, Optional (Keyword only)
            Whether the guild has the boost progress bar should be enabled.
        
        Returns
        -------
        guild : ``Guild`` object
            A partial guild made from the received data.
        
        Raises
        ------
        TypeError
            - If `icon` is neither `None`, `bytes-like`.
            - If `verification_level` was not given neither as ``VerificationLevel`` not `int`.
            - If `content_filter` was not given neither as ``ContentFilterLevel`` not `int`.
            - If `message_notification` was not given neither as ``MessageNotificationLevel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If the client cannot create more guilds.
            - If `name` was not given as `str`.
            - If the `name`'s length is out of range [2:100].
            - If `icon` is passed as `bytes-like`, but it's format is not a valid image format.
            - If `afk-timeout` was not given as `int`.
            - If `afk_timeout` was passed and not as one of: `60, 300, 900, 1800, 3600`.
            - If `boost_progress_bar_enabled` was not given as `bool`.
        """
        if __debug__:
            if self.bot:
                guild_create_limit = 10
            elif self.flags.staff:
                guild_create_limit = 200
            elif (self.premium_type is PremiumType.nitro):
                guild_create_limit = 200
            else:
                guild_create_limit = 100
            
            if len(self.guilds) >= guild_create_limit:
                raise AssertionError(
                    f'Guild count over creation limit; limit={guild_create_limit}; count={len(self.guilds)!r}.'
                )
        
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
            
            name_length = len(name)
            if name_length < 2 or name_length > 100:
                raise AssertionError(
                    f'`name` length can be in range [2:100], got {name_length!r}; {name!r}.'
                )
        
        
        if icon is None:
            icon_data = None
        else:
            icon_type = icon.__class__
            if not issubclass(icon_type, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`icon` can be `None`, `bytes-like`, got {icon_type.__name__}; {reprlib.repr(icon)}.'
                )
            
            if __debug__:
                media_type = get_image_media_type(icon)
                if media_type not in VALID_ICON_MEDIA_TYPES:
                    raise AssertionError(
                        f'Invalid `icon` type: {media_type}; {reprlib.repr(icon)}.'
                    )
            
            icon_data = image_to_base64(icon)
        
        if isinstance(verification_level, VerificationLevel):
            verification_level_value = verification_level.value
        elif isinstance(verification_level, int):
            verification_level_value = verification_level
        else:
            raise TypeError(
                f'`verification_level` can be `{VerificationLevel.__name__}`, `int`, got '
                f'{verification_level.__class__.__name__}; {verification_level!r}.'
            )
        
        
        if isinstance(message_notification, MessageNotificationLevel):
            message_notification_value = message_notification.value
        elif isinstance(message_notification, int):
            message_notification_value = message_notification
        else:
            raise TypeError(
                f'`message_notification` can be `{MessageNotificationLevel.__name__}`, `int`, got '
                f'{message_notification.__class__.__name__}; {message_notification!r}.'
            )
        
        
        if isinstance(content_filter, ContentFilterLevel):
            content_filter_value = content_filter.value
        elif isinstance(content_filter, int):
            content_filter_value = content_filter
        else:
            raise TypeError(
                f'`content_filter` can be {ContentFilterLevel.__name__}, `int` , got '
                f'{content_filter.__class__.__name__}; {content_filter!r}.'
            )
        
        if roles is None:
            roles = []
        
        if channels is None:
            channels = []
        
        data = {
            'name': name,
            'icon': icon_data,
            'verification_level': verification_level_value,
            'default_message_notifications': message_notification_value,
            'explicit_content_filter': content_filter_value,
            'roles': roles,
            'channels': channels,
        }
        
        if (afk_channel_id is not None):
            if __debug__:
                if not isinstance(afk_channel_id, int):
                    raise AssertionError(
                        f'`afk_channel_id` can be `int`, got {afk_channel_id.__class__.__name__}; {afk_channel_id!r}.'
                    )
            
            data['afk_channel_id'] = afk_channel_id
        
        if (system_channel_id is not None):
            if __debug__:
                if not isinstance(system_channel_id, int):
                    raise AssertionError(
                        f'`system_channel_id` can be `int`, got {system_channel_id.__class__.__name__}; '
                        f'{system_channel_id!r}.'
                    )
            
            data['system_channel_id'] = system_channel_id
        
        if (afk_timeout is not None):
            if __debug__:
                if not isinstance(afk_timeout, int):
                    raise AssertionError(
                        f'`afk_timeout` can be `int`, got {afk_timeout.__class__.__name__}; {afk_timeout!r}.'
                    )
                
                if afk_timeout not in (60, 300, 900, 1800, 3600):
                    raise AssertionError(
                        f'`afk_timeout` can be 60, 300, 900, 1800, 3600 seconds; got {afk_timeout!r}.'
                    )
            
            data['afk_timeout'] = afk_timeout
        
        if (boost_progress_bar_enabled is not None):
            if __debug__:
                if not isinstance(boost_progress_bar_enabled, bool):
                    raise AssertionError(
                        f'`boost_progress_bar_enabled` can be `bool`, got '
                        f'{boost_progress_bar_enabled.__class__.__name__}; {boost_progress_bar_enabled!r}.'
                    )
            
            data['premium_progress_bar_enabled'] = boost_progress_bar_enabled
        
        
        data = await self.http.guild_create(data)
        # we can create only partial, because the guild data is not completed usually
        return create_partial_guild_from_data(data)
    
    
    async def guild_prune(self, guild, days, *, roles=None, count=False, reason = None):
        """
        Kicks the members of the guild which were inactive since x days.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            Where the pruning will be executed.
        days : `int`
            The amount of days since at least the users need to inactive. Can be in range [1:30].
        roles : `None`, `list` of (``Role``, `int`) = `None`, Optional (Keyword only)
            By default pruning will kick only the users without any roles, but it can defined which roles to include.
        count : `bool` = `False`, Optional (Keyword only)
            Whether the method should return how much user were pruned, but if the guild is large it will be set to
            `False` anyways. Defaults to `False`.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the guild's audit logs.
        
        Returns
        -------
        count : `None`, `int`
            The number of pruned users or `None` if `count` is set to `False`.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `roles` contain not ``Role``, neither `int` element.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `roles` was given neither as `None`, `list`.
            - If `count` was not given as `bool`.
            - If `days` was not given as `int`.
            - If `days` is out of range [1:30].
        
        See Also
        --------
        ``.guild_prune_estimate`` : Returns how much user would be pruned if ``.guild_prune`` would be called.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        if __debug__:
            if not isinstance(count, bool):
                raise AssertionError(
                    f'`count` can be `bool`, got {count.__class__.__name__}; {count!r}.'
                )
        
        if count and (guild is not None) and guild.is_large:
            count = False
        
        if __debug__:
            if not isinstance(days, int):
                raise AssertionError(
                    f'``days` can be `int`, got {days.__class__.__name__}; {days!r}.'
                )
            
            if days < 1 or days > 30:
                raise AssertionError(
                    f'`days can be in range [1:30], got {days!r}.'
                )
        
        data = {
            'days': days,
            'compute_prune_count': count,
        }
        
        if (roles is not None):
            if __debug__:
                if not isinstance(roles, list):
                    raise AssertionError(
                        f'`roles` can be `None`m `list` of (`{Role.__name__}`, `int`)'
                        f', got {roles.__class__.__name__}; {roles!r}.')
            
            role_ids = set()
            for index, role in enumerate(roles):
                if isinstance(role, Role):
                    role_id = role.id
                else:
                    role_id = maybe_snowflake(role)
                    if role_id is None:
                        raise TypeError(
                            f'`roles[{index}]` is not `{Role.__name__}`, `int`,'
                            f' got {role.__class__.__name__}; {role!r}; roles={roles!r}.'
                        )
                
                role_ids.add(role_id)
            
            if role_ids:
                data['include_roles'] = role_ids
        
        data = await self.http.guild_prune(guild_id, data, reason)
        return data.get('pruned', None)
    
    
    async def guild_prune_estimate(self, guild, days, *, roles=None):
        """
        Returns the amount users, who would been pruned, if ``.guild_prune`` would be called.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`.
            Where the counting of prunable users will be done.
        days : `int`
            The amount of days since at least the users need to inactive. Can be in range [1:30].
        roles : `None`, `list` of (``Role``, `int`) = `None`, Optional (Keyword only)
            By default pruning would kick only the users without any roles, but it can be defined which roles to
            include.
        
        Returns
        -------
        count : `int`
            The amount of users who would be pruned.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `roles` contain not ``Role``, neither `int` element.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `roles` was given neither as `None`, `list`.
            - If `days` was not given as `int`.
            - If `days` is out of range [1:30].
        """
        guild_id = get_guild_id(guild)
        
        if __debug__:
            if not isinstance(days, int):
                raise AssertionError(f'`days can be `int`, got {days.__class__.__name__}.')
            
            if days < 1 or days > 30:
                raise AssertionError(f'`days can be in range [1:30], got {days!r}.')
        
        data = {
            'days': days,
        }
        
        if (roles is not None):
            if __debug__:
                if not isinstance(roles, list):
                    raise AssertionError(
                        f'`roles` can be `None`, `list` of (`{Role.__name__}`, `int`)'
                        f', got {roles.__class__.__name__}; {roles!r}.'
                    )
            
            role_ids = set()
            for role in roles:
                role_id = get_role_id(role)
                role_ids.add(role_id)
            
            if role_ids:
                data['include_roles'] = role_ids
        
        data = await self.http.guild_prune_estimate(guild_id, data)
        return data.get('pruned', None)
    
    
    async def guild_edit(
        self, guild, *, name=..., icon=..., invite_splash=..., discovery_splash=..., banner=...,
        afk_channel = ..., system_channel = ..., rules_channel = ..., public_updates_channel = ..., owner=...,
        afk_timeout = ..., verification_level=..., content_filter=..., message_notification=..., description = ...,
        preferred_locale=..., system_channel_flags=..., add_feature=..., remove_feature=...,
        boost_progress_bar_enabled=..., reason = None
    ):
        """
        Edits the guild with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to edit.
        name : `str`, Optional (Keyword only)
            The guild's new name.
        icon : `None`, `bytes-like`, Optional (Keyword only)
            The new icon of the guild. Can be `'jpg'`, `'png'`, `'webp'` image's raw data. If the guild has
            `ANIMATED_ICON` feature, it can be `'gif'` as well. By passing `None` you can remove the current one.
        invite_splash : `None`, `bytes-like`, Optional (Keyword only)
            The new invite splash of the guild. Can be `'jpg'`, `'png'`, `'webp'` image's raw data. The guild must have
            `INVITE_SPLASH` feature. By passing it as `None` you can remove the current one.
        discovery_splash : `None`, `bytes-like`, Optional (Keyword only)
            The new splash of the guild. Can be `'jpg'`, `'png'`, `'webp'` image's raw data. The guild must have
            `DISCOVERABLE` feature. By passing it as `None` you can remove the current one.
        banner : `None`, `bytes-like`, Optional (Keyword only)
            The new banner of the guild. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. The guild must
            have `BANNER` feature. `ANIMATED_BANNER` for `gif` banner. By passing it as `None` you can remove the
            current one.
        afk_channel : `None`, ``Channel``, `int`, Optional (Keyword only)
            The new afk channel of the guild. You can remove the current one by passing is as `None`.
        system_channel : `None`, ``Channel``, `int`, Optional (Keyword only)
            The new system channel of the guild. You can remove the current one by passing is as `None`.
        rules_channel : `None`, ``Channel``, `int`, Optional (Keyword only)
            The new rules channel of the guild. The guild must be a Community guild. You can remove the current
            one by passing is as `None`.
        public_updates_channel : `None`, ``Channel``, `int`, Optional (Keyword only)
            The new public updates channel of the guild. The guild must be a Community guild. You can remove the
            current one by passing is as `None`.
        owner : ``ClientUserBase``, `int`, Optional (Keyword only)
            The new owner of the guild. You must be the owner of the guild to transfer ownership.
            
        afk_timeout : `int`, Optional (Keyword only)
            The new afk timeout for the users at the guild's afk channel.
            
            Can be one of: `60, 300, 900, 1800, 3600`.
        verification_level : ``VerificationLevel``, `int`, Optional (Keyword only)
            The new verification level of the guild.
        content_filter : ``ContentFilterLevel``, `int`, Optional (Keyword only)
            The new content filter level of the guild.
        message_notification : ``MessageNotificationLevel``, Optional (Keyword only)
            The new message notification level of the guild.
        description : `None`, `str`, Optional (Keyword only)
            The new description of the guild. By passing `None`, or an empty string you can remove the current one.
        preferred_locale : ``Locale``, `str`, Optional (Keyword only)
            The guild's preferred locale. The guild must be a Community guild.
        system_channel_flags : ``SystemChannelFlag``, Optional (Keyword only)
            The guild's system channel's new flags.
        add_feature : (`str`, ``GuildFeature``) or (`iterable` of (`str`, ``GuildFeature``)), Optional (Keyword only)
            Guild feature(s) to add to the guild.
            
            If `guild` is given as an id, then `add_feature` should contain all the features of the guild to set.
        remove_feature : (`str`, ``GuildFeature``) or (`iterable` of (`str`, ``GuildFeature``)), Optional (Keyword only)
            Guild feature(s) to remove from the guild's.
        boost_progress_bar_enabled : `bool`, Optional (Keyword only)
            Whether the guild has the boost progress bar should be enabled.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, `str`.
            - If `icon`, `invite_splash`, `discovery_splash`, `banner` is neither `None`, `bytes-like`.
            - If `add_feature`, `remove_feature` was not given neither as `str`, ``GuildFeature``,
                `iterable` of `str`, ``GuildFeature``-s.
            - If `afk_channel` was given, but not as `None`, ``Channel``, neither as `int`.
            - If `system_channel`, `rules_channel`, `public_updates_channel` was given, but not as `None`,
                ``Channel``, neither as `int`.
            - If `owner` was not given neither as ``ClientUserBase``, `int`.
            - If `verification_level` was not given neither as ``VerificationLevel``, `int`.
            - If `content_filter` was not given neither as ``ContentFilterLevel``, `int`.
            - If `description` was not given either as `None`, `str`.
            - If `preferred_locale` was not given as ``Locale``, `str`.
        AssertionError
            - If `icon`, `invite_splash`, `discovery_splash`, `banner` was passed as `bytes-like`, but it's format
                is not any of the expected formats.
            - If `banner` was given meanwhile the guild has no `BANNER` feature.
            - If `rules_channel`, `preferred_locale`, `public_updates_channel` was passed meanwhile
                the guild is not Community guild.
            - If `owner` was passed meanwhile the client is not the owner of the guild.
            - If `afk_timeout` was passed and not as one of: `60, 300, 900, 1800, 3600`.
            - If `name` is shorter than `2` or longer than `100` characters.
            - If `discovery_splash` was given meanwhile the guild is not discoverable.
            - If `invite_splash` was passed meanwhile the guild has no `INVITE_SPLASH` feature.
            - If `name` was not given as `str`.
            - If `afk_timeout` was not given as `int`.
            - If `system_channel_flags` was not given as `SystemChannelFlag`, `int`.
            - If `boost_progress_bar_enabled` was not given as `bool`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = {}
        
        guild, guild_id = get_guild_and_id(guild)
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 2 or name_length > 100:
                    raise ValueError(
                        f'Guild\'s name\'s length can be between 2-100, got {name_length}: {name!r}.'
                    )
            
            data['name'] = name
        
        
        if (icon is not ...):
            if icon is None:
                icon_data = None
            
            else:
                if not isinstance(icon, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`icon` can be `None`, `bytes-like`, got {icon.__class__.__name__}; {reprlib.repr(icon)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(icon)
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid `icon` type for the guild: {media_type}; got {reprlib.repr(icon)}.'
                        )
                
                icon_data = image_to_base64(icon)
            
            data['icon'] = icon_data
        
        
        if (banner is not ...):
            if banner is None:
                banner_data = None
            else:
                if not isinstance(banner, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`banner` can be `None`, `bytes-like`, got '
                        f'{banner.__class__.__name__}; got {reprlib.repr(banner)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(banner)
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid `banner` type: {media_type}; got {reprlib.repr(banner)}.'
                        )
                
                banner_data = image_to_base64(banner)
            
            data['banner'] = banner_data
        
        
        if (invite_splash is not ...):
            if invite_splash is None:
                invite_splash_data = None
            else:
                if not isinstance(invite_splash, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`invite_splash` can be `None`, `bytes-like`, got '
                        f'{invite_splash.__class__.__name__}; {reprlib.repr(invite_splash)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(invite_splash)
                    if media_type not in VALID_ICON_MEDIA_TYPES:
                        raise AssertionError(
                            f'Invalid `invite_splash` type: {media_type}; got {reprlib.repr(invite_splash)}.'
                        )
                
                invite_splash_data = image_to_base64(invite_splash)
            
            data['splash'] = invite_splash_data
        
        
        if (discovery_splash is not ...):
            if discovery_splash is None:
                discovery_splash_data = None
            else:
                if not isinstance(discovery_splash, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`discovery_splash` can be `None`, `bytes-like`, got '
                        f'{discovery_splash.__class__.__name__}; {reprlib.repr(discovery_splash)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(discovery_splash)
                    if media_type not in VALID_ICON_MEDIA_TYPES:
                        raise AssertionError(
                            f'Invalid `discovery_splash` type: {media_type}; got {reprlib.repr(discovery_splash)}.'
                        )
                
                discovery_splash_data = image_to_base64(discovery_splash)
            
            data['discovery_splash'] = discovery_splash_data
        
        
        if (afk_channel is not ...):
            while True:
                if afk_channel is None:
                    afk_channel_id = None
                    break
                
                elif isinstance(afk_channel, Channel):
                    if afk_channel.is_guild_voice() or afk_channel.partial:
                        afk_channel_id = afk_channel.id
                        break
                
                else:
                    afk_channel_id = maybe_snowflake(afk_channel)
                    if afk_channel_id is not None:
                        break
                
                raise TypeError(
                    f'`afk_channel` can be `None`, guild voice channel, `int` , got '
                    f'{afk_channel.__class__.__name__}; {afk_channel!r}.'
                )
            
            data['afk_channel_id'] = afk_channel_id
        
        
        if (system_channel is not ...):
            while True:
                if system_channel is None:
                    system_channel_id = None
                    break
                
                elif isinstance(system_channel, Channel):
                    if system_channel.is_in_group_guild_system() or system_channel.partial:
                        system_channel_id = system_channel.id
                        break
                
                else:
                    system_channel_id = maybe_snowflake(system_channel)
                    if system_channel_id is not None:
                        break
                
                raise TypeError(
                    f'`system_channel` can be `None`, guild text channel, `int`, got '
                    f'{system_channel.__class__.__name__}; {system_channel!r}.'
                )
            
            data['system_channel_id'] = system_channel_id
        
        
        if (rules_channel is not ...):
            while True:
                if rules_channel is None:
                    rules_channel_id = None
                    break
                
                elif isinstance(rules_channel, Channel):
                    if rules_channel.is_in_group_guild_system() or rules_channel.partial:
                        rules_channel_id = rules_channel.id
                        break
                
                else:
                    rules_channel_id = maybe_snowflake(rules_channel)
                    if rules_channel_id is not None:
                        break
                
                raise TypeError(
                    f'`rules_channel` can be `None`, guild text channel, `int` , got '
                    f'{rules_channel.__class__.__name__}.'
                )
            
            data['rules_channel_id'] = rules_channel_id
        
        
        if (public_updates_channel is not ...):
            while True:
                if public_updates_channel is None:
                    public_updates_channel_id = None
                    break
                
                elif isinstance(public_updates_channel, Channel):
                    if public_updates_channel.is_in_group_guild_system() or public_updates_channel.partial:
                        public_updates_channel_id = public_updates_channel.id
                        break
                    
                else:
                    public_updates_channel_id = maybe_snowflake(public_updates_channel)
                    if public_updates_channel_id is not None:
                        break
                
                raise TypeError(
                    f'`public_updates_channel` can be `None`, guild text channel, `int`, got '
                    f'{public_updates_channel.__class__.__name__}; {public_updates_channel!r}.'
                )
            
            data['public_updates_channel_id'] = public_updates_channel_id
        
        
        if (owner is not ...):
            if __debug__:
                if (guild is not None) and (guild.owner_id != self.id):
                    raise AssertionError(
                        f'You must be owner to transfer ownership; got {owner!r}; actual owner={guild.owner!r}; '
                        f'guild={guild!r}.'
                    )
            
            if isinstance(owner, ClientUserBase):
                owner_id = owner.id
            else:
                owner_id = maybe_snowflake(owner)
                if owner_id is None:
                    raise TypeError(
                        f'`owner` can be `{UserBase.__name__}`, `int`, got {owner.__class__.__name__}; {owner!r}.'
                    )
            
            
            data['owner_id'] = owner_id
        
        if (afk_timeout is not ...):
            if __debug__:
                if not isinstance(afk_timeout, int):
                    raise AssertionError(
                        f'`afk_timeout` can be `int`, got {afk_timeout.__class__.__name__}; {afk_timeout!r}.'
                    )
                
                if afk_timeout not in (60, 300, 900, 1800, 3600):
                    raise AssertionError(
                        f'`afk_timeout` can be one of (60, 300, 900, 1800, 3600) seconds, got {afk_timeout!r}.'
                    )
            
            data['afk_timeout'] = afk_timeout
        
        
        if (verification_level is not ...):
            if isinstance(verification_level, VerificationLevel):
                verification_level_value = verification_level.value
            elif isinstance(verification_level, int):
                verification_level_value = verification_level
            else:
                raise TypeError(
                    f'`verification_level` can be `{VerificationLevel.__name__}`, `int`, got '
                    f'{verification_level.__class__.__name__}; {verification_level!r}.'
                )
            
            data['verification_level'] = verification_level_value
        
        
        if (content_filter is not ...):
            if isinstance(content_filter, ContentFilterLevel):
                content_filter_value = content_filter.value
            elif isinstance(content_filter, int):
                content_filter_value = content_filter
            else:
                raise TypeError(
                    f'`content_filter` can be `{ContentFilterLevel.__name__}`, `int`, got '
                    f'{content_filter.__class__.__name__}; {content_filter!r}.'
                )
            
            data['explicit_content_filter'] = content_filter_value
        
        
        if (message_notification is not ...):
            if isinstance(message_notification, MessageNotificationLevel):
                message_notification_value = message_notification.value
            elif isinstance(message_notification, int):
                message_notification_value = message_notification
            else:
                raise TypeError(
                    f'`message_notification` can be `{MessageNotificationLevel.__name__}`, `int`, got '
                    f'{message_notification.__class__.__name__}; {message_notification!r}.'
                )
            
            data['default_message_notifications'] = message_notification_value
        
        
        if (description is not ...):
            if description is None:
                pass
            elif isinstance(description, str):
                if not description:
                    description = None
            else:
                raise TypeError(
                    f'`description` can be `None`, `str`, got {description.__class__.__name__}; {description!r}.'
                )
            
            data['description'] = description
        
        
        if (preferred_locale is not ...):
            if isinstance(preferred_locale, Locale):
                preferred_locale_value = preferred_locale.value
            
            elif isinstance(preferred_locale, str):
                preferred_locale_value = preferred_locale
            
            else:
                raise TypeError(
                    f'`preferred_locale` can be `{Locale.__name__}`, `str`, got {preferred_locale.__class__.__name__}; '
                    f'{preferred_locale!r}.'
                )
            
            data['preferred_locale'] = preferred_locale_value
        
        
        if (system_channel_flags is not ...):
            if __debug__:
                if not isinstance(system_channel_flags, int):
                    raise AssertionError(
                        f'`system_channel_flags` can be `{SystemChannelFlag.__name__}`, `int`, got '
                        f'{system_channel_flags.__class__.__name__}; {system_channel_flags!r}.'
                    )
            
            data['system_channel_flags'] = system_channel_flags
        
        
        if (add_feature is not ...) or (remove_feature is not ...):
            # Collect actual
            features = set()
            if (guild is not None):
                for feature in guild.iter_features():
                    features.add(feature.value)
            
            # Collect added
            # Use GOTO
            while True:
                if add_feature is ...:
                    break
                
                if isinstance(add_feature, GuildFeature):
                    feature = add_feature.value
                elif isinstance(add_feature, str):
                    feature = add_feature
                else:
                    iter_func = getattr(type(add_feature), '__iter__', None)
                    if iter_func is None:
                        raise TypeError(
                            f'`add_feature` can be `str`, `{GuildFeature.__name__}`, `iterable` of '
                            f'(`str`, `{GuildFeature.__name__}`), got {add_feature.__class__.__name__}; '
                            f'{add_feature!r}.'
                        )
                    
                    for index, feature in enumerate(iter_func(add_feature)):
                        if isinstance(feature, GuildFeature):
                            feature = feature.value
                        elif isinstance(feature, str):
                            pass
                        else:
                            raise TypeError(
                                f'`add_feature` was given as `iterable` so it expected to have '
                                f'`{GuildFeature.__name__}`, `str` elements, but element `{index!r}` is '
                                f'{feature.__class__.__name__}; {feature!r}; add_feature={add_feature!r}.'
                            )
                        
                        features.add(feature)
                        continue
                    
                    break # End GOTO
                
                features.add(feature)
                break # End GOTO
            
            # Collect removed
            
            while True:
                if remove_feature is ...:
                    break
                
                if isinstance(remove_feature, GuildFeature):
                    feature = remove_feature.value
                elif isinstance(remove_feature, str):
                    feature = remove_feature
                else:
                    iter_func = getattr(type(remove_feature), '__iter__', None)
                    if iter_func is None:
                        raise TypeError(
                            f'`remove_feature` can be `str`, `{GuildFeature.__name__}`, `iterable` of '
                            f'(`str`, `{GuildFeature.__name__}`), got {remove_feature.__class__.__name__}; '
                            f'{remove_feature!r}.'
                        )
                    
                    for index, feature in enumerate(iter_func(remove_feature)):
                        if isinstance(feature, GuildFeature):
                            feature = feature.value
                        elif isinstance(feature, str):
                            pass
                        else:
                            raise TypeError(
                                f'`remove_feature` was given as `iterable` so it expected to have '
                                f'`{GuildFeature.__name__}`, `str` elements, but element `{index!r}` is '
                                f'{feature.__class__.__name__}; {feature!r}; remove_feature={remove_feature!r}.'
                            )
                        
                        features.discard(feature)
                        continue
                    
                    break # End GOTO
                
                features.discard(feature)
                break # End GOTO
            
            data['features'] = features
        
        
        if (boost_progress_bar_enabled is not ...):
            if __debug__:
                if not isinstance(boost_progress_bar_enabled, bool):
                    raise AssertionError(
                        f'`boost_progress_bar_enabled` can be `bool`, got '
                        f'{boost_progress_bar_enabled.__class__.__name__}; {boost_progress_bar_enabled!r}.'
                    )
            
            data['premium_progress_bar_enabled'] = boost_progress_bar_enabled
        
        
        await self.http.guild_edit(guild_id, data, reason)
    
    
    async def guild_widget_get(self, guild):
        """
        Returns the guild's widget.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild or the guild's id, what's widget will be requested.
        
        Returns
        -------
        guild_widget : `None`, ``GuildWidget``
            If the guild has it's widget disabled returns `None` instead.
        
        Raises
        ------
        TypeError
            If `guild` was not passed neither as ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        try:
            data = await self.http.guild_widget_get(guild_id)
        except DiscordException as err:
            if err.response.status == 403: # Widget Disabled -> return None
                return
            raise
        
        return GuildWidget(data)
    
    
    async def guild_user_get_all(self, guild):
        """
        Requests all the users of the guild and returns them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild what's users will be requested.
        
        Returns
        -------
        users : `list` of ``ClientUserBase`` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If user caching is allowed, these users should be already loaded if the client finished starting up.
        This method takes a long time to finish for huge guilds.
        
        When using it with user account, the client's token will be invalidated.
        """
        guild_id = get_guild_id(guild)
        
        data = {'limit': 1000, 'after': 0}
        users = []
        while True:
            guild_profile_datas = await self.http.guild_user_get_chunk(guild_id, data)
            for guild_profile_data in guild_profile_datas:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id)
                users.append(user)
            
            if len(guild_profile_datas) > 1000:
                break
            
            data['after'] = user.id
        
        return users
    
    
    async def guild_get_all(self):
        """
        Requests all the guilds of the client.
        
        This method is a coroutine.
        
        Returns
        -------
        guilds : `list` of ``Guild``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the client finished starting up, all the guilds should be already loaded.
        """
        result = []
        params = {'after': 0}
        while True:
            data = await self.http.guild_get_all(params)
            result.extend(create_partial_guild_from_data(guild_data) for guild_data in data)
            if len(data) < 100:
                break
            
            params['after'] = result[-1].id
        
        return result
    
    
    async def guild_voice_region_get_all(self, guild):
        """
        Requests the available voice regions for the given guild and returns them and the optional ones.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's regions will be requested.
        
        Returns
        -------
        voice_regions : `list` of ``VoiceRegion`` objects
            The available voice regions for the guild.
        optimals : `list` of ``VoiceRegion`` objects
            The optimal voice regions for the guild.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = await self.http.guild_voice_region_get_all(guild_id)
        voice_regions = []
        optimals = []
        for voice_region_data in data:
            region = VoiceRegion.from_data(voice_region_data)
            voice_regions.append(region)
            if voice_region_data['optimal']:
                optimals.append(region)
        
        return voice_regions, optimals
    
    
    async def voice_region_get_all(self):
        """
        Returns all the voice regions.
        
        This method is a coroutine.
        
        Returns
        -------
        voice_regions : `list` of ``VoiceRegion`` objects
            Received voice regions.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = await self.http.voice_region_get_all()
        voice_regions = []
        for voice_region_data in data:
            region = VoiceRegion.from_data(voice_region_data)
            voice_regions.append(region)
        
        return voice_regions
    
    
    async def audit_log_get_chunk(self, guild, limit=100, *, before=None, after=None, user=None, event = None):
        """
        Request a batch of audit logs of the guild and returns them. The `after`, `around` and the `before` parameters
        are mutually exclusive and they can be `int`, or as a ``DiscordEntity`` or as a `datetime`
        object.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's audit logs will be requested.
        limit : `int` = `100`, Optional
            The amount of audit logs to request. Can be between 1 and 100. Defaults to 100.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the audit log entries wer created.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the audit log entries wer created.
        user : `None`, ``ClientUserBase``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only to those, which were created by the given user.
        event : `None`, ``AuditLogEvent``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only on the given event.
        
        Returns
        -------
        audit_log : ``AuditLog``
            A container what contains the ``AuditLogEntry``-s.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, nor as `int`.
            - If `after`, `before` was passed with an unexpected type.
            - If `user` is neither `None`, ``ClientUserBase``, `int`.
            - If `event` is neither `None`, ``AuditLogEvent``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `int`.
            - If `limit` is out of the expected range [1:100].
        """
        guild, guild_id = get_guild_and_id(guild)
        
        if __debug__:
            if not isinstance(limit, int):
                raise AssertionError(
                    f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}.'
                )
            
            if limit < 1 or limit > 100:
                raise ValueError(
                    f'`limit` out of the expected range [1:100], got {limit!r}.'
                )
        
        data = {'limit': limit}
        
        if (before is not None):
            data['before'] = log_time_converter(before)
        
        if (after is not None):
            data['after'] = log_time_converter(after)
        
        if (user is not None):
            if isinstance(user, ClientUserBase):
                user_id = user.id
            
            else:
                user_id = maybe_snowflake(user)
                if user_id is None:
                    raise TypeError(
                        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
                    )
            
            data['user_id'] = user_id
        
        if (event is not None):
            if isinstance(event, AuditLogEvent):
                event_value = event.value
            elif isinstance(event, int):
                event_value = event
            else:
                raise TypeError(
                    f'`event` can be `None`, `{AuditLogEvent.__name__}`, `int`, got '
                    f'{event.__class__.__name__}; {event!r}.'
                )
            
            data['action_type'] = event_value
        
        data = await self.http.audit_log_get_chunk(guild_id, data)
        if guild is None:
            guild = create_partial_guild_from_id(guild_id)
        
        return AuditLog(data, guild)
    
    
    async def audit_log_iterator(self, guild, *, user=None, event = None):
        """
        Returns an audit log iterator for the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's audit logs will be requested.
        user : `None`, ``ClientUserBase``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only to those, which were created by the given user.
        event : `None`, ``AuditLogEvent``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only on the given event.
        
        Returns
        -------
        audit_log_iterator : ``AuditLogIterator``
        """
        return await AuditLogIterator(self, guild, user=user, event = event)



__all__ = ()

from scarletio import Compound, IgnoreCaseMultiValueDictionary
from scarletio.web_common import BasicAuth
from scarletio.web_common.headers import AUTHORIZATION

from ...application import Application, ApplicationRoleConnection
from ...application.application_role_connection.utils import APPLICATION_ROLE_CONNECTION_FIELD_CONVERTERS
from ...bases import maybe_snowflake
from ...guild import create_partial_guild_from_data
from ...http import DiscordHTTPClient
from ...oauth2 import Connection, Oauth2Access, Oauth2Scope, Oauth2User
from ...oauth2.helpers import build_joined_scopes, join_oauth2_scopes
from ...payload_building import build_edit_payload
from ...role import Role

from ..request_helpers import get_guild_id, get_oauth2_access_token, get_oauth2_access_token_and_user_id
from ..utils import UserGuildPermission

from .application_command import _assert__application_id


def _assert__redirect_url(redirect_url):
    """
    Asserts whether the `redirect_url`'s type is correct.
    
    Parameters
    ----------
    redirect_url : `str`
        The url, where the activation page redirected to.
    
    Raises
    ------
    AssertionError
        - If `redirect_url` is not `str`.
    """
    if not isinstance(redirect_url, str):
        raise AssertionError(
            f'`redirect_url` can be `str`, got {redirect_url.__class__.__name__}; {redirect_url!r}.'
        )
    
    return True


def _assert__code(code):
    """
    Asserts whether the `code`'s type is correct.
    
    Parameters
    ----------
    code : `str`
        The code, what is included with the redirect url after a successful activation.
    
    Raises
    ------
    AssertionError
        - If `code` is not `str`.
    """
    if not isinstance(code, str):
        raise AssertionError(
            f'`code` can be `str`, got {code.__class__.__name__}; {code!r}.'
        )
    
    return True


class ClientCompoundOauth2Endpoints(Compound):
    
    application : Application
    http : DiscordHTTPClient
    id : int
    secret : str
    
    
    async def activate_authorization_code(self, redirect_url, code, scopes):
        """
        Activates a user's oauth2 code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        redirect_url : `str`
            The url, where the activation page redirected to.
        code : `str`
            The code, what is included with the redirect url after a successful activation.
        scopes : `str`, `list` of `str`
            Scope or a  list of oauth2 scopes to request.
        
        Returns
        -------
        access : ``Oauth2Access``, `None`
            If the code, the redirect url or the scopes are invalid, the methods returns `None`.
        
        Raises
        ------
        TypeError
            If `Scopes` wasn't neither as `str` not `list` of `str`s.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``parse_oauth2_redirect_url`` : Parses `redirect_url` and the `code` from a full url.
        """
        assert _assert__redirect_url(redirect_url)
        assert _assert__code(code)
        
        joined_scopes = build_joined_scopes(scopes)
        
        data = {
            'client_id': self.id,
            'client_secret': self.secret,
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': redirect_url,
            'scope': joined_scopes,
        }
        
        data = await self.http.oauth2_token(data, IgnoreCaseMultiValueDictionary())
        if len(data) == 1:
            return
        
        return Oauth2Access(data, redirect_url)
    
    
    async def owners_access(self, scopes):
        """
        Similar to ``.activate_authorization_code``, but it requests the application's owner's access. It does not
        requires the redirect_url and the code parameter either.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scopes : `list` of `str`
            A list of oauth2 scopes to request.
        
        Returns
        -------
        access : ``Oauth2Access``
            The oauth2 access of the client's application's owner.
        
        Raises
        ------
        TypeError
            If `Scopes` is neither `str` nor `list` of `str`s.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `redirect_url` was not given as `str`.
            - If `code` was not given as `str`.
            - If `scopes` is empty.
            - If `scopes` contains empty string.
        
        Notes
        -----
        Does not work if the client's application is owned by a team.
        """
        joined_scopes = build_joined_scopes(scopes)
        
        data = {
            'grant_type': 'client_credentials',
            'scope': joined_scopes,
        }
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = BasicAuth(str(self.id), self.secret).encode()
        data = await self.http.oauth2_token(data, headers)
        return Oauth2Access(data, '')
    
    
    async def user_info_get(self, access):
        """
        Request the a user's information with oauth2 access token. By default a bot account should be able to request
        every public information about a user (but you do not need oauth2 for that). If the access token has email
        or/and identify scopes, then more information should show up like this.
        
        This method is a coroutine.
        
        Parameters
        ----------
        access : ``Oauth2Access``, ``Oauth2User``, `str`
            Oauth2 access to the respective user or it's access token.
        
        Returns
        -------
        oauth2_user : ``Oauth2User``
            The requested user object.
        
        Raises
        ------
        TypeError
            - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Needs `'email'` or / and `'identify'` scopes granted for more data
        """
        access_token = get_oauth2_access_token(access)
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        data = await self.http.user_info_get(headers)
        return Oauth2User(data, access)
        
    
    async def user_connection_get_all(self, access):
        """
        Requests a user's connections. This method will work only if the access token has the `'connections'` scope. At
        the returned list includes the user's hidden connections as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        access : ``Oauth2Access``, ``Oauth2User``, `str`
            Oauth2 access to the respective user or it's access token.
        
        Returns
        -------
        connections : `list` of ``Connection``
            The user's connections.
        
        Raises
        ------
        TypeError
            - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        ValueError
            - If the given `access` is not providing the required scope.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        access_token = get_oauth2_access_token(access, Oauth2Scope.connections)
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        data = await self.http.user_connection_get_all(headers)
        return [Connection.from_data(connection_data) for connection_data in data]
    
    
    async def renew_access_token(self, access):
        """
        Renews the access token of an ``Oauth2Access``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        access : ``Oauth2Access``, ``Oauth2User``
            Oauth2 access to the respective user.
        
        Raises
        ------
        TypeError
            - If `access` is not ``Oauth2Access``, ``Oauth2User``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        By default access tokens expire after one week.
        """
        if not isinstance(access, (Oauth2Access, Oauth2User)):
            raise TypeError(
                f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`'
                f', got {access.__class__.__name__}; {access!r}.'
            )
        
        redirect_url = access.redirect_url
        if redirect_url:
            data = {
                'client_id': self.id,
                'client_secret': self.secret,
                'grant_type': 'refresh_token',
                'refresh_token': access.refresh_token,
                'redirect_uri': redirect_url,
                'scope': join_oauth2_scopes(access.scopes)
            }
        else:
            data = {
                'client_id': self.id,
                'client_secret': self.secret,
                'grant_type': 'client_credentials',
                'scope': join_oauth2_scopes(access.scopes),
            }
        
        data = await self.http.oauth2_token(data, IgnoreCaseMultiValueDictionary())
        
        access._renew(data)
    
    
    async def guild_user_add(self, guild, access, user=None, *, nick=None, roles=None, mute=False, deaf=False):
        """
        Adds the passed to the guild. The user must have granted you the `'guilds.join'` oauth2 scope.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, where the user is going to be added.
        
        access: ``Oauth2Access``, ``Oauth2User``, `str`
            The access of the user, who will be added.
        
        user : `None`, ```ClientUserBase`` = `None`, `int`, Optional
            Defines which user will be added to the guild. The `access` must refer to this specified user.
            
            This field is optional if access is passed as an ``Oauth2User`` object.
        
        nick : `None`, `str` = `None`, Optional (Keyword only)
            The nickname, which with the user will be added.
        
        roles : `None`, `list` of (``Role``, `int`) = `None`, Optional (Keyword only)
            The roles to add the user with.
        
        mute : `bool` = `False`, Optional (Keyword only)
            Whether the user should be added as muted.
        
        deaf : `bool` = `False`, Optional (Keyword only)
            Whether the user should be added as deafen.
        
        Raises
        ------
        TypeError:
            - If `user` was not given neither as `None`, ``ClientUserBase``, `int`.
            - If `user` was passed as `None` and `access` was passed as ``Oauth2Access``, `str`.
            - If `access` was not given as ``Oauth2Access``, ``Oauth2User``, nether as `str`.
            - If the given `access` not grants `'guilds.join'` scope.
            - If `guild` was not given neither as ``Guild``, not `int`.
            - If `roles` contain not ``Role``, nor `int`.
        ValueError
            - If the given `access` is not providing the required scope.
            - If `user` and `access` refers to a different user.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        access_token, user_id = get_oauth2_access_token_and_user_id(access, user, Oauth2Scope.guilds_join)
        guild_id = get_guild_id(guild)
        
        
        data = {'access_token': access_token}
        
        
        # Security debug checks.
        if __debug__:
            if (nick is not None):
                if not isinstance(nick, str):
                    raise AssertionError(
                        f'`nick` can be `None`, `str`, got {nick.__class__.__name__}; {nick!r}.'
                    )
                
                nick_length = len(nick)
                if nick_length > 32:
                    raise AssertionError(
                        f'`nick` length can be in range [0:32], got {nick_length}; {nick!r}.'
                    )
        
        if (nick is not None) and nick:
            data['nick'] = nick
        
        
        if (roles is not None):
            if __debug__:
                if not isinstance(roles, list):
                    raise AssertionError(
                        f'`roles` can be `list` of (`{Role.__name__}`, `int`), got '
                        f'{roles.__class__.__name__}; {roles!r}.'
                    )
            
            if roles:
                role_ids = set()
                
                for index, role in enumerate(roles):
                    if isinstance(role, Role):
                        role_id = role.id
                    else:
                        role_id = maybe_snowflake(role)
                        if role_id is None:
                            raise TypeError(
                                f'`roles[{index}]` is neither `{Role.__name__}`, `int`, got '
                                f'{role.__class__.__name__}; {role!r}; roles={roles!r}.'
                            )
                    
                    role_ids.add(role_id)
                
                data['roles'] = role_ids
        
        
        if __debug__:
            if not isinstance(mute, bool):
                raise AssertionError(
                    f'`mute` can be `bool`, got {mute.__class__.__name__}; {mute!r}.'
                )
        
        if mute:
            data['mute'] = mute
        
        
        if __debug__:
            if not isinstance(deaf, bool):
                raise AssertionError(
                    f'`deaf` can be `bool`, got {deaf.__class__.__name__}; {deaf!r}.'
                )
        
        if deaf:
            data['deaf'] = deaf
        
        
        await self.http.guild_user_add(guild_id, user_id, data)
    
    
    async def user_guild_get_all(self, access):
        """
        Requests a user's guilds with it's ``Oauth2Access``.
        The user must provide the `Oauth2Scope.guilds` scope for this request to succeed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        access: ``Oauth2Access``, ``Oauth2User``, `str`
            The access of the user, who's guilds will be requested.
        
        Returns
        -------
        guilds_and_permissions : `list` of `tuple` (``Guild``, ``UserGuildPermission``)
            The guilds and the user's permissions in each of them. Not loaded guilds will show up as partial ones.
        
        Raises
        ------
        TypeError
            - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        ValueError
            - If the given `access` is not providing the required scope.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        access_token = get_oauth2_access_token(access, Oauth2Scope.guilds)
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        data = await self.http.user_guild_get_all(headers)
        return [(create_partial_guild_from_data(guild_data), UserGuildPermission(guild_data)) for guild_data in data]
    
    
    async def user_application_role_connection_get(self, access):
        """
        Requests a user's application role connections with it's ``Oauth2Access``.
        The user must provide the `Oauth2Scope.role_connections_write` scope for this request to succeed.
                
        This method is a coroutine.
        
        Parameters
        ----------
        access: ``Oauth2Access``, ``Oauth2User``, `str`
            The access of the user, who will's application role connections will be requested..
        
        Returns
        -------
        application_role_connection : ``ApplicationRoleConnection``
        
        Raises
        ------
        TypeError
            - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        ValueError
            - If the given `access` is not providing the required scope.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        access_token = get_oauth2_access_token(access, Oauth2Scope.role_connections_write)
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        data = await self.http.user_application_role_connection_get(application_id, headers)
        return ApplicationRoleConnection.from_data(data)
    
    
    async def user_application_role_connection_edit(
        self, access, application_role_connection_template = None, **keyword_parameters
    ):
        """
        Edits a user's application role connections with it's ``Oauth2Access``.
        The user must provide the `Oauth2Scope.role_connections_write` scope for this request to succeed.
                
        This method is a coroutine.
        
        Parameters
        ----------
        access: ``Oauth2Access``, ``Oauth2User``, `str`
            The access of the user, who will's application role connections will be requested..
        
        application_role_connection_template : `None`, ``ApplicationRoleConnection`` = `None`, Optional
            Application role connection to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        platform_name : `None`, `str`, Optional (Keyword only)
            The vanity name of the platform the application represents.
        platform_user_name : `None`, `str`, Optional (Keyword only)
            The name of the user on the application's platform.
        metadata_values : `None`, `dict` of (`str`, `str`) items, Optional (Keyword only)
            Metadata key to attached value relation.
        
        Returns
        -------
        application_role_connection : ``ApplicationRoleConnection``
        
        Raises
        ------
        TypeError
            - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        ValueError
            - If the given `access` is not providing the required scope.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        access_token = get_oauth2_access_token(access, Oauth2Scope.role_connections_write)
        
        data = build_edit_payload(
            None,
            application_role_connection_template,
            APPLICATION_ROLE_CONNECTION_FIELD_CONVERTERS,
            keyword_parameters,
        )
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        data = await self.http.user_application_role_connection_edit(application_id, data, headers)
        return ApplicationRoleConnection.from_data(data)



__all__ = ('ApplicationRoleConnectionMetadata',)

from scarletio import RichAttributeErrorBaseType

from ...localization.utils import hash_locale_dictionary

from .fields import (
    parse_description, parse_description_localizations, parse_key, parse_name, parse_name_localizations, parse_type,
    put_description_into, put_description_localizations_into, put_key_into, put_name_into, put_name_localizations_into,
    put_type_into, validate_description, validate_description_localizations, validate_key, validate_name,
    validate_name_localizations, validate_type
)
from .helpers import escape_name_to_key
from .preinstanced import ApplicationRoleConnectionMetadataType


class ApplicationRoleConnectionMetadata(RichAttributeErrorBaseType):
    """
    Role connection metadata of an application.
    
    If a bot's application has `.role_connection_verification_url` configured, then the application will show up as a
    verification method in the guild's settings.
    
    These metadatas will appear in the role verification configuration when the application has been linked to a role.
    
    When a user connects their account using the bot's `.role_connection_verification_url`, the bot will update the
    user's role connection with metadata using the oauth2 `role_connections.write` scope.
    
    Attributes
    ----------
    description : `None`, `str`
        The metadata's description.
    description_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized descriptions of the metadata.
    key : `str`
        The dictionary key for the metadata.
    name : `str`
        The name of the metadata.
    name_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized names of the metadata.
    type : ``ApplicationRoleConnectionMetadataType``
        The dictionary value's type and their respective operation.
    """
    __slots__ = ('description', 'description_localizations', 'key', 'name', 'name_localizations', 'type')
    
    def __new__(
        cls,
        name,
        metadata_type, 
        *,
        description = ...,
        description_localizations = ...,
        key = ...,
        name_localizations = ...,
    ):
        """
        Creates a new role connection metadata.
        
        Parameters
        ----------
        name : `str`
            The name of the metadata.
        
        metadata_type : `int`, ``ApplicationRoleConnectionMetadataType``
            The dictionary value's type and their respective operation.
        
        description : `None`, `str`, Optional (Keyword only)
            The metadata's description.
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items, Optional (Keyword only)
            Localized descriptions of the metadata.
        
        key : `str`, Optional (Keyword only)
            The dictionary key for the metadata.
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items, Optional (Keyword only)
            Localized names of the metadata.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # Required: name & metadata_type
        name = validate_name(name)
        metadata_type = validate_type(metadata_type)
        
        # description
        if description is ...:
            description = name
        else:
            description = validate_description(description)
        
        # description_localizations
        if description_localizations is ...:
            description_localizations = None
        else:
            description_localizations = validate_description_localizations(description_localizations)
        
        # key
        if key is ...:
            key = escape_name_to_key(name)
        else:
            key = validate_key(key)

        # name_localizations
        if name_localizations is ...:
            name_localizations = None
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        self = object.__new__(cls)
        self.description = description
        self.description_localizations = description_localizations
        self.key = key
        self.name = name
        self.name_localizations = name_localizations
        self.type = metadata_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a role connection metadata from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Role connection data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.description = parse_description(data)
        self.description_localizations = parse_description_localizations(data)
        self.key = parse_key(data)
        self.name = parse_name(data)
        self.name_localizations = parse_name_localizations(data)
        self.type = parse_type(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the role connection metadata to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_description_into(self.description, data, defaults)
        put_description_localizations_into(self.description_localizations, data, defaults)
        put_key_into(self.key, data, defaults)
        put_name_into(self.name, data, defaults)
        put_name_localizations_into(self.name_localizations, data, defaults)
        put_type_into(self.type, data, defaults)
        return data
    
    
    def __hash__(self):
        """Returns the role connection metadata's hash value."""
        hash_value = 0
        
        # description
        description = self.description
        if (description is not None) and (description != self.name):
            hash_value ^= hash(description)
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            hash_value ^= hash_locale_dictionary(description_localizations)
        
        # key
        key = self.key
        if (key != self.key):
            hash_value ^= hash(key)
        
        # name
        hash_value ^= hash(self.name)
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            hash_value ^= hash_locale_dictionary(name_localizations)
        
        # type
        hash_value ^= self.type.value
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the role connection metadata's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        metadata_type = self.type
        repr_parts.append(', type = ')
        repr_parts.append(str(metadata_type.name))
        repr_parts.append('~')
        repr_parts.append(repr(metadata_type.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # description
        if self.description != other.description:
            return False
        
        # description_localizations
        if self.description_localizations != other.description_localizations:
            return False
        
        # key
        if self.key != other.key:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # name_localizations
        if self.name_localizations != other.name_localizations:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the role connection metadata.
        
        Returns
        -------
        new : `instance<type<cls>>`
        """
        new = object.__new__(type(self))
        new.description = self.description
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            description_localizations = description_localizations.copy()
        new.description_localizations = description_localizations
        new.key = self.key
        new.name = self.name
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            name_localizations = name_localizations.copy()
        new.name_localizations = name_localizations
        new.type = self.type
        return new
    
    
    def copy_with(
        self,
        *,
        description = ...,
        description_localizations = ...,
        key = ...,
        metadata_type = ...,
        name = ...,
        name_localizations = ...,
    ):
        """
        Copies the role connection metadata with the given fields.
        
        Parameters
        ----------
        description : `None`, `str`, Optional (Keyword only)
            The metadata's description.
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items, Optional (Keyword only)
            Localized descriptions of the metadata.
        
        key : `str`, Optional (Keyword only)
            The dictionary key for the metadata.
        
        metadata_type : `int`, ``ApplicationRoleConnectionMetadataType``, Optional (Keyword only)
            The dictionary value's type and their respective operation.
        
        name : `str`, Optional (Keyword only)
            The name of the metadata.
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items, Optional (Keyword only)
            Localized names of the metadata.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # description_localizations
        if description_localizations is ...:
            description_localizations = self.description_localizations
            if (description_localizations is not None):
                 description_localizations = description_localizations.copy()
        else:
            description_localizations = validate_description_localizations(description_localizations)
        
        # key
        if key is ...:
            key = self.key
        else:
            key = validate_key(key)
        
        # metadata_type
        if metadata_type is ...:
            metadata_type = self.type
        else:
            metadata_type = validate_type(metadata_type)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # name_localizations
        if name_localizations is ...:
            name_localizations = self.name_localizations
            if (name_localizations is not None):
                 name_localizations = name_localizations.copy()
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        
        new = object.__new__(type(self))
        new.description = description
        new.description_localizations = description_localizations
        new.key = key
        new.name = name
        new.name_localizations = name_localizations
        new.type = metadata_type
        return new



__all__ = ('ScheduledEventEntityMetadataLocation',)

from scarletio import copy_docs

from .base import ScheduledEventEntityMetadataBase


class ScheduledEventEntityMetadataLocation(ScheduledEventEntityMetadataBase):
    """
    Location entity metadata of ``ScheduledEvent``-s.
    
    Attributes
    ----------
    location : `None`, `str`
        The place where the event will take place.
    """
    __slots__ = ('location', )
    
    def __new__(cls, location):
        """
        Creates a new location entity metadata for ``ScheduledEvent``-s.
        
        Parameters
        ----------
        location : `str`
            The location.
        
        Raises
        ------
        TypeError
            If `location`'s type is incorrect.
        ValueError
            If `location` is an empty string.
        """
        if not isinstance(location, str):
            raise TypeError(
                f'`location` can be `str`, got {location.__class__.__name__}; {location!r}.'
            )
        
        if not location:
            raise ValueError(
                f'`location` cannot be empty string.'
            )
        
        self = object.__new__(cls)
        self.location = location
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadataBase.from_data)
    def from_data(cls, data):
        location = data.get('location', None)
        
        self = object.__new__(cls)
        self.location = location
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadataBase.to_data)
    def to_data(self):
        return {
            'location': self.location,
        }
    
    
    @copy_docs(ScheduledEventEntityMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.location != other.location:
            return False
        
        return True
    
    @copy_docs(ScheduledEventEntityMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        location = self.location
        if (location is not None):
            hash_value ^= hash(location)
        
        return hash_value



__all__ = ('ActivityMetadataRich',)

import warnings
from datetime import datetime

from scarletio import copy_docs

from ...preconverters import preconvert_flag, preconvert_snowflake, preconvert_str
from ...utils import datetime_to_millisecond_unix_time, is_url, millisecond_unix_time_to_datetime

from ..fields import ActivityAssets, ActivityParty, ActivitySecrets, ActivityTimestamps
from ..flags import ActivityFlag

from .base import ActivityMetadataBase


class ActivityMetadataRich(ActivityMetadataBase):
    """
    Represents a Discord rich activity.
    
    Attributes
    ----------
    application_id : `int`
        The id of the activity's application. Defaults to `0`.
    assets : `None`, ``ActivityAssets``
        The activity's assets. Defaults to `None`.
    created_at : `None`, `datetime`
        When the activity was created. Defaults to Discord epoch.
    details : `None`, `str`
        What the player is currently doing. Defaults to `None`.
    flags : ``ActivityFlag``
        The flags of the activity. Defaults to `ActivityFlag(0)`
    id : `int`
        The id of the activity. Defaults to `0`.
    name : `str`
        The activity's name.
    party : `None`, ``ActivityParty``
        The activity's party.
    secrets : `None`, ``ActivitySecrets``
        The activity's secrets. Defaults to `None`.
    session_id : `None`, `str`
        Spotify activity's session's id. Defaults to `None`.
    state : `None`, `str`
        The player's current party status. Defaults to `None`.
    sync_id : `None`, `str`
        The ID of the currently playing track of a spotify activity. Defaults to `None`.
    timestamps : `None`, ``ActivityTimestamps``
        The activity's timestamps.
    url : `None`, `str`
        The url of the stream (Twitch or Youtube only). Defaults to `None`.
    """
    __slots__ = (
        'application_id', 'assets', 'created_at', 'details', 'flags', 'id', 'name', 'party', 'secrets', 'session_id',
        'state', 'sync_id', 'timestamps', 'url'
    )
    
    @copy_docs(ActivityMetadataBase.__new__)
    def __new__(cls, keyword_parameters):
        # application_id
        try:
            application_id = keyword_parameters.pop('application_id')
        except KeyError:
            application_id = 0
        else:
            application_id = preconvert_snowflake(application_id, 'application_id')
        
        # assets
        try:
            assets = keyword_parameters.pop('assets')
        except KeyError:
            assets = None
        else:
            if (assets is not None) and (not isinstance(assets, ActivityAssets)):
                raise TypeError(
                    f'`assets` can be `None`, `{ActivityAssets.__name__}`, got '
                    f'{assets.__class__.__name__}; {assets!r}.'
                )
        
        # created_at
        try:
            created_at = keyword_parameters.pop('created_at')
        except KeyError:
            created_at = None
        else:
            if (created_at is not None) and (not isinstance(created_at, datetime)):
                raise TypeError(
                    f'`created_at` can be `datetime`, got {created_at.__class__.__name__}; {created_at!r}.'
                )
        
        # details
        try:
            details = keyword_parameters.pop('details')
        except KeyError:
            details = None
        else:
            if (details is not None):
                details = preconvert_str(details, 'details', 0, 2048)
                if (not details):
                    details = None
        
        # flags
        try:
            flags = keyword_parameters.pop('flags')
        except KeyError:
            flags = ActivityFlag()
        else:
            flags = preconvert_flag(flags, 'flags', ActivityFlag)
        
        try:
            activity_id = keyword_parameters.pop('id_')
        except KeyError:
            pass
        else:
            warnings.warn(
                (
                    f'`{cls.__name__}.__new__`\'s `id_` parameter is deprecated and will be removed in 2023 Marc. '
                    f'Please use `activity_id` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            keyword_parameters['activity_id'] = activity_id
        
        # activity_id
        try:
            activity_id = keyword_parameters.pop('activity_id')
        except KeyError:
            activity_id = 0
        else:
            activity_id = preconvert_snowflake(activity_id, 'activity_id')
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            name = ''
        else:
            name = preconvert_str(name, 'name', 1, 2048)
        
        # party
        try:
            party = keyword_parameters.pop('party')
        except KeyError:
            party = None
        else:
            if (party is not None) and (not isinstance(party, ActivityParty)):
                raise TypeError(
                    f'`party` can be `None`, `{ActivityParty.__name__}`, got '
                    f'{party.__class__.__name__}; {party!r}.'
                )
        
        # secrets
        try:
            secrets = keyword_parameters.pop('secrets')
        except KeyError:
            secrets = None
        else:
            if (secrets is not None) and (not isinstance(secrets, ActivitySecrets)):
                raise TypeError(
                    f'`secrets` can be `None`, `{ActivitySecrets.__name__}`, got '
                    f'{secrets.__class__.__name__}; {secrets!r}.'
                )
        
        # session_id
        try:
            session_id = keyword_parameters.pop('session_id')
        except KeyError:
            session_id = None
        else:
            if (session_id is not None):
                session_id = preconvert_str(session_id, 'session_id', 0, 2048)
                if (not session_id):
                    session_id = None
        
        # state
        try:
            state = keyword_parameters.pop('state')
        except KeyError:
            state = None
        else:
            if (state is not None):
                state = preconvert_str(state, 'state', 0, 2048)
                if (not state):
                    state = None
        
        # timestamps
        try:
            timestamps = keyword_parameters.pop('timestamps')
        except KeyError:
            timestamps = None
        else:
            if (timestamps is not None) and (not isinstance(timestamps, ActivityTimestamps)):
                raise TypeError(
                    f'`timestamps` can be `None`, `{ActivityTimestamps.__name__}`, got '
                    f'{timestamps.__class__.__name__}; {timestamps!r}.'
                )
        
        # sync_id
        try:
            sync_id = keyword_parameters.pop('sync_id')
        except KeyError:
            sync_id = None
        else:
            if (sync_id is not None):
                sync_id = preconvert_str(sync_id, 'sync_id', 0, 2048)
                if (not sync_id):
                    sync_id = None
        
        # url
        try:
            url = keyword_parameters.pop('url')
        except KeyError:
            url = None
        else:
            if (url is not None):
                url = preconvert_str(url, 'url', 0, 2048)
                if url:
                    assert is_url(url), f'`url` was not given as a valid url, got {url!r}.'
                else:
                    url = None
        
        
        self = object.__new__(cls)
        self.name = name
        
        self.application_id = application_id
        self.details = details
        self.flags = flags
        self.state = state
        self.party = party
        self.assets = assets
        self.secrets = secrets
        self.sync_id = sync_id
        self.session_id = session_id
        self.created_at = created_at
        self.id = activity_id
        self.timestamps = timestamps
        self.url = url
        
        return self
    
    
    @copy_docs(ActivityMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # application_id
        hash_value ^= self.application_id
        
        # assets
        assets = self.assets
        if (assets is not None):
            hash_value ^= hash(assets)
        
        # created_at
        created_at = self.created_at
        if (created_at is not None):
            hash_value ^= hash(created_at)
        
        # details
        details = self.details
        if (details is not None):
            hash_value ^= hash(details)
        
        # flags
        hash_value ^= self.flags
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # party
        party = self.party
        if (party is not None):
            hash_value ^= hash(party)
        
        # secrets
        secrets = self.secrets
        if (secrets is not None):
            hash_value ^= hash(secrets)
        
        # session_id
        session_id = self.session_id
        if (session_id is not None):
            hash_value ^= hash(session_id)
        
        # state
        state = self.state
        if (state is not None):
            hash_value ^= hash(state)
        
        # sync_id
        sync_id = self.sync_id
        if (sync_id is not None):
            hash_value ^= hash(sync_id)
        
        # timestamps
        timestamps = self.timestamps
        if (timestamps is not None):
            hash_value ^= hash(timestamps)
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= hash(url)
        
        return hash_value
    
    
    @copy_docs(ActivityMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # assets
        if self.assets != other.assets:
            return False
        
        # created_at
        if self.created_at != other.created_at:
            return False
        
        # details
        if self.details != other.details:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # party
        if self.party != other.party:
            return False
        
        # secrets
        if self.secrets != other.secrets:
            return False
        
        # session_id
        if self.session_id != other.session_id:
            return False
        
        # state
        if self.state != other.state:
            return False
        
        # sync_id
        if self.sync_id != other.sync_id:
            return False
        
        # timestamps
        if self.timestamps != other.timestamps:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ActivityMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        # application_id
        application_id = data.get('application_id', None)
        if (application_id is None):
            application_id = 0
        else:
            application_id = int(application_id)
        self.application_id = application_id
        
        # id
        try:
            raw_id = data['id']
        except KeyError:
            id_ = 0
        else:
            try:
                id_ = int(raw_id, base=16)
            except ValueError:
                # Some activity types have `custom` id-s, at those cases we default back to `0`.
                id_ = 0
        
        self.id = id_
        
        self._update_attributes(data)
        return self
    
    
    @copy_docs(ActivityMetadataBase.to_data)
    def to_data(self, *, include_internals=False, user=False):
        data = {}
        
        # name
        data['name'] = self.name
        
        # url
        url = self.url
        if (url is not None):
            data['url'] = url
        
        if user or include_internals:
            # assets
            assets = self.assets
            if (assets is not None) and assets:
                data['assets'] = assets.to_data()
            
            # details
            details = self.details
            if (details is not None):
                data['details'] = details
            
            # party
            party = self.party
            if (party is not None) and party:
                data['party'] = party.to_data()
            
            # secrets
            secrets = self.secrets
            if (secrets is not None) and secrets:
                data['secrets'] = secrets.to_data()
            
            # state
            state = self.state
            if (state is not None):
                data['state'] = state
            
            # timestamps
            timestamps = self.timestamps
            if (timestamps is not None) and timestamps:
                data['timestamps'] = timestamps.to_data()
        
        
        if include_internals:
            # application_id | receive only?
            application_id = self.application_id
            if application_id:
                data['application_id'] = str(application_id)
            
            # created_at | receive only?
            created_at = self.created_at
            if (created_at is not None):
                data['created_at'] = datetime_to_millisecond_unix_time(created_at)
            
            # flags | spotify only
            flags = self.flags
            if flags:
                data['flags'] = int(flags)
            
            # session_id | spotify only
            session_id = self.session_id
            if (session_id is not None):
                data['session_id'] = session_id
            
            # sync_id | spotify only
            sync_id = self.sync_id
            if (sync_id is not None):
                data['sync_id'] = sync_id
        
        return data
    
    
    @copy_docs(ActivityMetadataBase._update_attributes)
    def _update_attributes(self, data):
        # application_id
        # never changes
        
        # assets
        assets_data = data.get('assets', None)
        if (assets_data is None):
            assets = None
        else:
            assets = ActivityAssets.from_data(assets_data)
        self.assets = assets
        
        # created_at
        created_at = data.get('created_at', None)
        if (created_at is not None):
            created_at = millisecond_unix_time_to_datetime(created_at)
        self.created_at = created_at
        
        # details
        self.details = data.get('details', None)
        
        # flags
        self.flags = ActivityFlag(data.get('flags', 0))
        
        # id
        # Never changes
        
        # name
        self.name = data.get('name', '')
        
        # party
        party_data = data.get('party', None)
        if (party_data is None):
            party = None
        else:
            party = ActivityParty.from_data(party_data)
        self.party = party
        
        # secrets
        secrets_data = data.get('secrets', None)
        if (secrets_data is None):
            secrets = None
        else:
            secrets = ActivitySecrets.from_data(secrets_data)
        self.secrets = secrets
        
        # session_id
        self.session_id = data.get('session_id', None)
        
        # state
        self.state = data.get('state', None)
        
        # sync_id
        self.sync_id = data.get('sync_id', None)
        
        # timestamps
        timestamps_data = data.get('timestamps', None)
        if timestamps_data is None:
            timestamps = None
        else:
            timestamps = ActivityTimestamps.from_data(timestamps_data)
        self.timestamps = timestamps
        
        # url
        self.url = data.get('url', None)
    
    
    @copy_docs(ActivityMetadataBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = {}
        
        # application_id
        application_id = data.get('application_id', None)
        if (application_id is None):
            application_id = 0
        else:
            application_id = int(application_id)
        
        if self.application_id != application_id:
            old_attributes['application_id'] = self.application_id
            self.application_id = application_id
        
        # assets
        assets_data = data.get('assets', None)
        if (assets_data is None):
            assets = None
        else:
            assets = ActivityAssets.from_data(assets_data)
        
        if self.assets != assets:
            old_attributes['assets'] = self.assets
            self.assets = assets
        
        # created_at
        created_at = data.get('created_at', None)
        if created_at is not None:
            created_at = millisecond_unix_time_to_datetime(created_at)
        
        if self.created_at != created_at:
            old_attributes['created_at'] = self.created_at
            self.created_at = created_at
        
        # details
        details = data.get('details', None)
        if self.details != details:
            old_attributes['details'] = self.details
            self.details = details
        
        # flags
        flags = data.get('flags', 0)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = ActivityFlag(flags)
        
        # name
        name = data.get('name', '')
        
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # party
        party_data = data.get('party', None)
        if (party_data is None):
            party = None
        else:
            party = ActivityParty.from_data(party_data)
        
        if self.party != party:
            old_attributes['party'] = self.party
            self.party = party
        
        # secrets
        secrets_data = data.get('secrets', None)
        if (secrets_data is None):
            secrets = None
        else:
            secrets = ActivitySecrets.from_data(secrets_data)
        
        if self.secrets != secrets:
            old_attributes['secrets'] = self.secrets
            self.secrets = secrets
        
        # session_id
        session_id = data.get('session_id', None)
        if self.session_id != session_id:
            old_attributes['session_id'] = self.session_id
            self.session_id = session_id
        
        # state
        state = data.get('state', None)
        
        if self.state != state:
            old_attributes['state'] = self.state
            self.state = state
        
        # sync_id
        sync_id = data.get('sync_id', None)
        if self.sync_id != sync_id:
            old_attributes['sync_id'] = self.sync_id
            self.sync_id = sync_id
        
        # timestamps
        timestamps_data = data.get('timestamps', None)
        if timestamps_data is None:
            timestamps = None
        else:
            timestamps = ActivityTimestamps.from_data(timestamps_data)
        
        if self.timestamps != timestamps:
            old_attributes['timestamps'] = self.timestamps
            self.timestamps = timestamps
        
        # url
        url = data.get('url', None)
        if self.url != url:
            old_attributes['url'] = self.url
            self.url = url
        
        return old_attributes



__all__ = ('thread_user_create', 'thread_user_delete', 'thread_user_difference_update', 'thread_user_pop')

from scarletio import include

from .thread_profile import ThreadProfile

Client = include('Client')


def thread_user_create(thread_channel, user, thread_profile_data):
    """
    Resolves the given thread profile data.
    
    Parameters
    ----------
    thread_channel : ``Channel``
        The respective thread.
    user : ``ClientUserBase``
        The respective user to add or update in the thread.
    thread_profile_data : `dict` of (`str`, `Any`) items
        Received thread profile data.
    
    Returns
    -------
    created : `bool`
        Whether a new thread profile was created.
    """
    thread_users = thread_channel.thread_users
    if thread_users is None:
        thread_users = thread_channel.thread_users = {}
    thread_users[user.id] = user
    
    thread_profiles = user.thread_profiles
    if thread_profiles is None:
        thread_profiles = user.thread_profiles = {}
    
    try:
        thread_profile = thread_profiles[thread_channel.id]
    except KeyError:
        thread_profiles[thread_channel.id] = ThreadProfile.from_data(thread_profile_data)
        created = True
    else:
        thread_profile._update_attributes(thread_profile_data)
        created = False
    
    return created


def thread_user_difference_update(thread_channel, user, thread_profile_data):
    """
    Resolves the given thread profile update.
    
    Parameters
    ----------
    thread_channel : ``Channel``
        The respective thread.
    user : ``ClientUserBase``
        The respective user to add or update in the thread.
    thread_profile_data : `dict` of (`str`, `Any`) items
        Received thread profile data.
    
    Returns
    -------
    old_attributes : `None`, `dict` of (`str`, `Any`) items
    """
    thread_users = thread_channel.thread_users
    if thread_users is None:
        thread_users = thread_channel.thread_users = {}
    thread_users[user.id] = user
    
    thread_profiles = user.thread_profiles
    if thread_profiles is None:
        thread_profiles = user.thread_profiles = {}
    
    try:
        thread_profile = thread_profiles[thread_channel.id]
    except KeyError:
        thread_profiles[thread_channel.id] = ThreadProfile.from_data(thread_profile_data)
        return None
    
    old_attributes = thread_profile._difference_update_attributes(thread_profile_data)
    if not old_attributes:
        old_attributes = None
    
    return old_attributes


def thread_user_delete(thread_channel, user_id):
    """
    Removes the user for the given id from the thread's users.
    
    Parameters
    ----------
    thread_channel : ``Channel``
        The respective thread.
    user_id : `int`
        The respective user's identifier.
    """
    thread_users = thread_channel.thread_users
    if (thread_users is not None):
        try:
            user = thread_users.pop(user_id)
        except KeyError:
            pass
        else:
            if not thread_users:
                thread_channel.thread_users = None
            
            thread_profiles = user.thread_profiles
            if (thread_profiles is not None):
                try:
                    del thread_profiles[thread_channel.id]
                except KeyError:
                    pass
                else:
                    if not thread_profiles:
                        user.thread_profiles = None


def thread_user_pop(thread_channel, user_id, me):
    """
    Removes and returns the user for the given id from the thread's users.
    
    Parameters
    ----------
    thread_channel : ``Channel``
        The respective thread.
    user_id : `int`
        The respective user's identifier.
    me : ``Client``
        The client who pops the user.
    
    Returns
    -------
    popped : `None`, `tuple` (``ClientUserBase``, ``ThreadProfile``) item
        The removed user and it's profile if any.
    """
    thread_users = thread_channel.thread_users
    if (thread_users is not None):
        try:
            user = thread_users[user_id]
        except KeyError:
            pass
        else:
            if not thread_users:
                thread_channel.thread_users = None
            
            thread_profiles = user.thread_profiles
            if (thread_profiles is not None):
                if isinstance(user, Client) and (user is not me):
                    thread_profile = thread_profiles.get(thread_channel.id, None)
                else:
                    try:
                        thread_profile = thread_profiles.pop(thread_channel.id)
                    except KeyError:
                        thread_profile = None
                    else:
                        if not thread_profiles:
                            user.thread_profiles = None
                    
                    # at the above condition we dont wanna pop obviously.
                    try:
                        del thread_users[user_id]
                    except KeyError:
                        pass
                    else:
                        if not thread_users:
                            thread_channel.thread_users = None
                
                if (thread_profile is not None):
                    return user, thread_profile



__all__ = ('ComponentMetadataChannelSelect', )

from scarletio import copy_docs

from .fields import parse_channel_types, put_channel_types_into, validate_channel_types
from .select_base import ComponentMetadataSelectBase


class ComponentMetadataChannelSelect(ComponentMetadataSelectBase):
    """
    Channel select component metadata.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool`
        Whether the component is enabled.
    
    max_values : `int
        The maximal amount of options to select.
    
    min_values : `int`
        The minimal amount of options to select.
    
    placeholder : `None`, `str`
        Placeholder text of the select.
    
    channel_types : `None`, `tuple` of ``ChannelType``
        The allowed channel types by the select.
    """
    __slots__ = ('channel_types',)

    @copy_docs(ComponentMetadataSelectBase.__new__)
    def __new__(cls, keyword_parameters):
        # channel_types
        try:
            channel_types = keyword_parameters.pop('channel_types')
        except KeyError:
            channel_types = None
        else:
            channel_types = validate_channel_types(channel_types)
        
        # Construct
        self = ComponentMetadataSelectBase.__new__(cls, keyword_parameters)
        self.channel_types = channel_types
        
        return self
    
    
    @copy_docs(ComponentMetadataSelectBase._add_type_specific_repr_fields_into)
    def _add_type_specific_repr_fields_into(self, repr_parts):
        # channel_types
        repr_parts.append(', channel_types = ')
        channel_types = self.channel_types
        if (channel_types is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(channel_types)
            
            while True:
                option = channel_types[index]
                index += 1
                
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
    
    
    @copy_docs(ComponentMetadataSelectBase.__hash__)
    def __hash__(self):
        hash_value = ComponentMetadataSelectBase.__hash__(self)
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            hash_value ^= len(channel_types) << 12
            for index, channel_type in enumerate(channel_types):
                hash_value ^= channel_type.value << (index << 1)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataSelectBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ComponentMetadataSelectBase._is_equal_same_type(self, other):
            return False
        
        # channel_types
        if self.channel_types != other.channel_types:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataSelectBase.from_data)
    def from_data(cls, data):
        self = super(ComponentMetadataChannelSelect, cls).from_data(data)
        self.channel_types = parse_channel_types(data)
        return self
    
    
    @copy_docs(ComponentMetadataSelectBase.to_data)
    def to_data(self, *, defaults = False):
        data =  ComponentMetadataSelectBase.to_data(self)
        
        put_channel_types_into(self.channel_types, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataSelectBase.copy)
    def copy(self):
        new = ComponentMetadataSelectBase.copy(self)
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            channel_types = tuple(channel_types)
        new.channel_types = channel_types
        
        return new
    
    
    @copy_docs(ComponentMetadataSelectBase.copy_with)
    def copy_with(self, keyword_parameters):
        # channel_types
        try:
            channel_types = keyword_parameters.pop('channel_types')
        except KeyError:
            channel_types = self.channel_types
            if (channel_types is not None):
                channel_types = tuple(channel_types)
        else:
            channel_types = validate_channel_types(channel_types)
        
        # Construct
        
        new = ComponentMetadataSelectBase.copy_with(self, keyword_parameters)
        new.channel_types = channel_types
        return new



__all__ = ()

from ...field_parsers import (
    default_entity_parser_factory, preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    default_entity_putter_factory, preinstanced_array_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    default_entity_validator, preinstanced_array_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

from .preinstanced import TeamMemberPermission, TeamMembershipState

# permissions

parse_permissions = preinstanced_array_parser_factory('permissions', TeamMemberPermission)
put_permissions_into = preinstanced_array_putter_factory('permissions')
validate_permissions = preinstanced_array_validator_factory('permissions', TeamMemberPermission)

# state

parse_state = preinstanced_parser_factory('membership_state', TeamMembershipState, TeamMembershipState.none)
put_state_into = preinstanced_putter_factory('membership_state')
validate_state = preinstanced_validator_factory('state', TeamMembershipState)

# user

parse_user = default_entity_parser_factory('user', User, ZEROUSER)
put_user_into = default_entity_putter_factory('user', ClientUserBase, ZEROUSER)
validate_user = default_entity_validator('user', ClientUserBase, ZEROUSER)



__all__ = ()

from ...field_parsers import preinstanced_parser_factory
from ...field_putters import preinstanced_putter_factory
from ...field_validators import preinstanced_validator_factory

from ..action_metadata import AutoModerationActionMetadataBase

from .preinstanced import AutoModerationActionType

# type

parse_type = preinstanced_parser_factory('type', AutoModerationActionType, AutoModerationActionType.none)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', AutoModerationActionType)

# metadata

def parse_metadata(data, action_type):
    """
    Parsers out an auto moderation action's metadata form the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Auto moderation action data.
    
    action_type : ``AutoModerationActionType``
        The action's type.
    
    Returns
    -------
    metadata : ``AutoModerationActionMetadataBase``
    """
    metadata_type = action_type.metadata_type
    metadata_data = data.get('metadata', None)
    
    if metadata_data is None:
        metadata = metadata_type()
    else:
        metadata = metadata_type.from_data(metadata_data)
    
    return metadata


def put_metadata_into(metadata, data, defaults):
    """
    Puts the given action metadata's data into the given `data` json serializable object.
    
    Parameters
    ----------
    metadata : ``AutoModerationActionMetadataBase``
        Action metadata.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if defaults or (type(metadata) is not AutoModerationActionMetadataBase):
        data['metadata'] = metadata.to_data(defaults = defaults)
    
    return data



__all__ = ('OperationSystem',)

from ...bases import Preinstance as P, PreinstancedBase


class OperationSystem(PreinstancedBase):
    """
    Represents a ``ApplicationExecutable``'s operation system.
    
    Attributes
    ----------
    name : `str`
        The name of state.
    value : `int`
        The Discord side identifier value of the os.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``Os``) items
        Stores the created os instances. This container is accessed when translating a Discord
        os' value to it's representation.
    VALUE_TYPE : `type` = `int`
        The os' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the oss.
    
    Every predefined os can be accessed as class attribute as well:
    +-----------------------+-----------+---------------+
    | Class attribute name  | name      | value         |
    +=======================+===========+===============+
    | none                  | none      | `''`          |
    +-----------------------+-----------+---------------+
    | darwin                | darwin    | '`darwin'`    |
    +-----------------------+-----------+---------------+
    | linux                 | linux     | '`linux`'     |
    +-----------------------+-----------+---------------+
    | windows               | windows   | `'win32'`     |
    +-----------------------+-----------+---------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ()
    
    # predefined
    none = P('', 'none')
    darwin = P('darwin', 'darwin')
    linux = P('linux', 'linux')
    windows = P('win32', 'windows')



__all__ = ('SolarPlayer', )

from datetime import datetime
from random import randrange

from scarletio import copy_docs

from ...discord.core import GUILDS
from ...discord.utils import datetime_to_timestamp

from .filters import Volume
from .player_base import SolarPlayerBase
from .track import ConfiguredTrack, Track


class SolarPlayer(SolarPlayerBase):
    """
    Player of solar link.
    
    Attributes
    ----------
    _forward_data : `None`, `dict` of (`str`, `Any`) items
        Json to forward to the player's node as necessary.
    _position : `float`
        The position of the current track.
    _position_update : `float`
        When ``._position`` was last updated in monotonic time.
    channel_id : `int`
        The channel's identifier, where the node is connected to.
    guild_id : `int`
        The guild's identifier, where the player is.
    node : ``SolarNode``
        The node that the player is connected to. Defaults to `None`.
    _current_track : `None`, ``ConfiguredTrack``
        The currently played track if any.
    _paused : `bool`
        Whether the player is paused.
    _paused_track : `None`, ``ConfiguredTrack``
        The paused track.
    _repeat_current : `bool`
        Whether the current track should be repeated. Defaults to `False`.
    _repeat_queue : `bool`
        Whether tracks should be repeated.
    _shuffle : `bool`
        Whether tracks should be shuffled.
    queue : `list` of ``ConfiguredTrack``
        The queue of the tracks to play.
    """
    __slots__ = (
        '_current_track', '_paused', '_paused_track', '_repeat_current', '_repeat_queue', '_shuffle', 'queue'
    )
    
    @copy_docs(SolarPlayerBase.__new__)
    def __new__(cls, node, guild_id, channel_id):
        """
        Creates a new solar player instance.
        
        Parameters
        ----------
        node : ``SolarNode``
            The node of the player.
        guild_id : `int`
            The guild's identifier, where the node will connect to.
        channel_id : `int`
            The channel's identifier, where the node will connect to.
        
        Returns
        -------
        waiter : ``Future`` to ``SolarPlayerBase``
            A future, which can be awaited to get the connected player.
        """
        self, waiter = SolarPlayerBase.__new__(cls, node, guild_id, channel_id)
        
        self._shuffle = False
        self._repeat_queue = False
        self._repeat_current = False
        self._current_track = None
        self._paused = None
        self._paused_track = None
        self.queue = []
        
        return self, waiter
    
    
    @copy_docs(SolarPlayerBase.is_paused)
    def is_paused(self):
        return self._paused
    
    
    @copy_docs(SolarPlayerBase.get_current)
    def get_current(self):
        return self._current_track
    
    
    def get_volume(self):
        """
        Returns the player's volume.
        
        Returns
        -------
        volume : `float`
        """
        filter = self.get_filter(Volume)
        if filter is None:
            volume = 1.0
        else:
            volume = filter._volume
        
        return volume
    
    
    async def append(self, track, start_time=0.0, end_time=0.0, **added_attributes):
        """
        Adds a new track to play.
        
        This method is a coroutine.
        
        Parameters
        ----------
        track : ``Track``
            The track to play.
        start_time : `float` = `0.0`, Optional
            Where the track will start in seconds.
        end_time : `float` = `0.0`, Optional
            Where the track will start in seconds.
        **added_attributes : `dict` of (`str`, `Any`)
            Additional user defined attributes.
        
        Returns
        -------
        started_playing : `bool`
            Whether the source is started playing and not put on queue.
        
        Raises
        ------
        TypeError
            If `track` is neither ``Track``, nor ``ConfiguredTrack``.
        ValueError
            - If `start_time` is out of the expected [0.0:duration] range.
            - If `end_time` is out of the expected [0.0:duration] range.
        """
        configured_track = ConfiguredTrack(track, start_time, end_time, added_attributes)
        
        if self._current_track is None:
            paused = self._paused
            if (not paused):
                await self._play(configured_track)
            
            self._current_track = configured_track
            
            if paused:
                started_playing = False
            else:
                started_playing = True
        else:
            self.queue.append(configured_track)
            started_playing = False
        
        return started_playing
    
    
    def move_track(self, source_index, target_index):
        """
        Moves the track to the target index.
        
        Parameters
        ----------
        source_index : `int`
            The track's current index.
        target_index : `int`
            the track's new index.
        
        Returns
        -------
        track : `None`, ``ConfiguredTrack``
            The moved track if any.
        
        Raises
        ------
        AssertionError
            - If `source_index` is not `int`.
            - If `target_index` is not `int`.
        """
        if __debug__:
            if not isinstance(source_index, int):
                raise AssertionError(
                    f'`source_index` can be `int`, got {source_index.__class__.__name__}; {source_index!r}.'
                )
            
            if not isinstance(target_index, int):
                raise AssertionError(
                    f'`target_index` can be `int`, got {target_index.__class__.__name__}; {target_index!r}.'
                )
        
        queue = self.queue
        if queue:
            queue_length = len(queue)
            
            if source_index < 0:
                source_index = 0
            elif source_index >= queue_length:
                source_index = queue_length-1
            
            if target_index < 0:
                target_index = 0
            
            
            if target_index == source_index:
                track = None
            
            else:
                track = queue.pop(source_index)
                queue.insert(target_index, track)
        
        else:
            track = None
        
        return track
    
    
    def iter_all_track(self):
        """
        Iterates over all the tracks of the player.
        
        This method is an iterable generator.
        
        Yields
        ------
        track : ``ConfiguredTrack``
        """
        current_track = self._current_track
        if (current_track is not None):
            yield current_track
        
        yield from self.queue
    
    
    async def stop(self):
        """
        Stops the player's current track.
        
        This method is a coroutine.
        """
        # Are we already stopped?
        if self._current_track is None:
            return
        
        await self._stop()
        
        # do not change `self._paused`
        self._paused_track = None
        self._current_track = None
    
    
    async def skip(self, index=0):
        """
        Skips the given track by it's index and returns it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        index : `int` = `0.0`, Optional
            The track's index to skip.
            
            When skipping the `0`-th track, so the current, it will start to play the next if not paused.
        
        Returns
        -------
        track : `None`, ``ConfiguredTrack``
        """
        queue = self.queue
        
        if index == 0:
            track = self._current_track
            self._current_track = None
            
            if self._repeat_current:
                if not self._paused:
                    await self._play(track)
                
                self._current_track = track
            
            elif self._repeat_queue:
                # Only repeat the actual, if there is no other.
                if queue:
                    if self._shuffle:
                        track_index = randrange(len(queue))
                    else:
                        track_index = 0
                    
                    new_track = queue[track_index]
                else:
                    new_track = track
                    track_index = -1
                
                if not self._paused:
                    await self._play(new_track)
                
                self._current_track = new_track
                
                if track_index != -1:
                    del queue[track_index]
                    queue.append(track)
            
            else:
                if queue:
                    if self._shuffle:
                        track_index = randrange(len(queue))
                    else:
                        track_index = 0
                    new_track = queue[track_index]
                else:
                    new_track = None
                    track_index = -1
                
                if new_track is None:
                    await self._stop()
                
                else:
                    if not self._paused:
                        await self._play(new_track)
                    
                    self._current_track = new_track
                
                if track_index != -1:
                    del queue[track_index]
        
        elif index < 0:
            track = None
        
        else:
            if index > len(queue):
                track = None
            else:
                track = queue.pop(index - 1)
                
                if self._repeat_queue and (not self._repeat_current):
                    queue.append(track)
        
        return track
    
    
    async def remove(self, index=0):
        """
        Removes the track with the given index from the queue and returns it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        index : `int` = `0.0`, Optional
            The track's index to skip.
            
            When skipping the `0`-th track, so the current, it will start to play the next if not paused.
        
        Returns
        -------
        track : `None`, ``ConfiguredTrack``
        """
        queue = self.queue
        
        if index == 0:
            track = self._current_track
            self._current_track = None
            
            if queue:
                if self._shuffle:
                    track_index = randrange(len(queue))
                else:
                    track_index = 0
                new_track = queue[track_index]
                
                if not self._paused:
                    await self._play(new_track)
                
                self._current_track = new_track
                del queue[index]
            
            else:
                if (track is not None):
                    await self._stop()
        
        elif index > len(queue):
            track = None
        else:
            track = queue.pop(index - 1)
        
        return track
    
    
    async def pause(self):
        """
        Pauses the player.
        
        This method is a coroutine.
        """
        if self._paused:
            return
        
        current_track = self._current_track
        if (current_track is not None):
            await self._pause()
        
        self._paused = True
        self._paused_track = current_track
    
    
    async def resume(self):
        """
        Resumes the player.
        
        This method is a coroutine.
        """
        if not self._paused:
            return
        
        paused_track = self._paused_track
        if (paused_track is not None):
            current_track = self._current_track
            if (current_track is not None):
                node = self.node
                if (node is not None):
                    if (paused_track is current_track):
                        await self._resume()
                    else:
                        await self._play(current_track)
        
        self._paused_track = None
        self._paused = False
    
    
    def set_repeat_queue(self, repeat):
        """
        Sets the player to repeat the whole queue.
        
        Parameters
        ----------
        repeat : `bool`
            Whether to repeat queue.
        """
        self._repeat_queue = repeat
        
        if repeat:
            self._repeat_current = False
    
    
    def set_repeat_current(self, repeat):
        """
        Sets the player to repeat the actual track.
        
        Parameters
        ----------
        repeat : `bool`
            Whether to repeat the player or not.
        actual : `bool`
            Whether the repeat the current track.
        """
        self._repeat_current = repeat
        
        if repeat:
            self._repeat_queue = False
    
    
    def is_repeating_queue(self):
        """
        Returns whether the player is repeating the queue.
        
        Returns
        -------
        repeat : `bool`
        """
        return self._repeat_queue
    
    
    def is_repeating_current(self):
        """
        Returns whether the player is repeating the current track.
        
        Returns
        -------
        repeat_current : `bool`
        """
        return self._repeat_current
    
    
    def set_shuffle(self, shuffle):
        """
        Sets the player's shuffle the tracks.
        
        Note, that if the current actual track is repeated shuffling is ignored.
        
        Parameters
        ----------
        shuffle : `bool`
            Whether to shuffle the player or not.
        """
        self._shuffle = shuffle
    
    
    def is_shuffling(self):
        """
        Returns whether the player is shuffling the tracks.
        
        Returns
        -------
        shuffle : `bool`
        """
        return self._shuffle
    
    
    async def set_volume(self, volume):
        """
        Sets the player's volume.
        
        This method is a coroutine.
        
        Parameters
        ----------
        volume : `float`
            The new volume level. Can be in range [0.0:10.0].
        """
        self.add_filter(Volume(volume))
        await self.apply_filters()
    
    
    async def seek(self, position):
        """
        Seeks to a given position in the track.
        
        This method is a coroutine.
        
        Parameters
        ----------
        position : `float`
            The new position in seconds.
        """
        if (self._current_track is None):
            return
        
        await self._seek(position)
    
    
    async def join_speakers(self, *, request=False):
        """
        Requests to speak at the voice client's voice channel. Only applicable for stage channels.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : `bool` = `False`, Optional (Keyword only)
            Whether the client should only request to speak.
        
        Raises
        ------
        RuntimeError
            Operation on disconnected player.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        node = self.node
        if node is None:
            # Already disconnected.
            return
        
        guild_id = self.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            try:
                voice_state = guild.voice_states[node.client.id]
            except KeyError:
                return
            
            if voice_state.is_speaker:
                return
        
        if request:
            timestamp = datetime_to_timestamp(datetime.utcnow())
        else:
            timestamp = None
        
        data = {
            'suppress': False,
            'request_to_speak_timestamp': timestamp,
            'channel_id': self.channel_id
        }
        
        await node.client.http.voice_state_client_edit(guild_id, data)
    
    
    async def join_audience(self):
        """
        Joins the audience in the voice client's voice channel. Only applicable for stage channels.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        node = self.node
        if node is None:
            # Already disconnected.
            return
        
        guild_id = self.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            try:
                voice_state = guild.voice_states[node.client.id]
            except KeyError:
                return
            
            if not voice_state.is_speaker:
                return
        
        data = {
            'suppress': True,
            'channel_id': self.channel_id
        }
        
        await node.client.http.voice_state_client_edit(guild_id, data)



from .component import *
from .component_metadata import *
from .interaction_form import *
from .string_select_option import *

from .shared_constants import *
from .shared_fields import *
from .shared_helpers import *
from .utils import *


__all__ = (
    *component.__all__,
    *component_metadata.__all__,
    *interaction_form.__all__,
    *string_select_option.__all__,
    
    *shared_constants.__all__,
    *shared_fields.__all__,
    *shared_helpers.__all__,
    *utils.__all__,
)


from ...utils.module_deprecation import deprecated_import
deprecated_import(StringSelectOption, 'ComponentSelectOption')
deprecated_import(Component, 'ComponentBase')
deprecated_import(create_button, 'ComponentButton')
deprecated_import(create_row, 'ComponentRow')
deprecated_import(create_string_select, 'ComponentSelect')
deprecated_import(create_text_input, 'ComponentTextInput')
